import os
import re
import shutil
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
    final,
    overload,  # noqa: F401
)

from typing_extensions import TypeAlias, override

from lazy_type_hint.file_modifiers.py_file_modifier import PyFileModifier
from lazy_type_hint.file_modifiers.yaml_file_modifier import YAML_COMMENTS_POSITION
from lazy_type_hint.generators.lazy_type_hint_abc import LazyTypeHintABC
from lazy_type_hint.strategies import ParsingStrategies
from lazy_type_hint.utils import (
    TAB,
    is_string_python_keyword_compatible,
)

THIS_DIR = Path(__file__).parent


class LazyTypeHintLiveError(Exception):
    """Raised by `LazyTypeHintLive` class."""


ObjectT = TypeVar("ObjectT")
PathT = TypeVar("PathT", str, Path)


class LazyTypeHintLive(LazyTypeHintABC):
    """Generate type hints for any data structure and use them out of the box."""

    # Utils
    this_file_pyi: PyFileModifier
    """.pyi representation of this same module."""
    strategies: ParsingStrategies
    """Strategies to follow when parsing the objects."""
    if_type_hint_exists: Literal["overwrite", "validate"]
    """Strategy to follow if existing type hints were found for the same class name."""

    # Constants that should not be modified
    classes_created: "TypeAlias" = Any
    """Classes created by the class. Do not modify."""
    _methods_to_be_overloaded: Final = ("from_data", "from_yaml_file")
    """Methods that will be modified in the PYI interface when new classes are added."""
    _this_file_pyi_path: Final = Path(__file__).with_suffix(".pyi")
    """Path to the .pyi file associated to this same module."""
    _custom_class_dir_path: Final = _this_file_pyi_path.parent / "build"
    """Path where the new classes will be generated.c"""

    @final
    def __init__(
        self,
        strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
        if_type_hint_exists: Literal["overwrite", "validate"] = "validate",
    ) -> None:
        self.strategies = strategies
        self.if_type_hint_exists = if_type_hint_exists

        if not self._this_file_pyi_path.exists():
            self.this_file_pyi = self._generate_this_file_pyi()
        else:
            self.this_file_pyi = PyFileModifier(self._this_file_pyi_path.read_text(encoding="utf-8"))

    def _get_classes_added(self) -> Mapping[str, Path]:
        to_find = "classes_created"
        values = self.this_file_pyi.search_assignment("classes_created", only_values=True)
        if not values:
            raise LazyTypeHintLiveError(f"No `{to_find}` was found in this file.")

        if values[0] == "Any":
            return {}
        matches = re.findall(r'"(.*?)"', values[0])

        dct: Dict[str, Path] = {}
        for match in matches:
            path = Path(self._custom_class_dir_path / f"{match}.py")
            if not path.exists():
                raise LazyTypeHintLiveError(
                    f"A class `{match}` was apparently created but cannot find its "
                    f"corresponding source code within {self._custom_class_dir_path}"
                )
            dct[match] = path
        return dct

    @override
    def from_yaml_file(
        self,
        loader: Callable[[PathT], ObjectT],
        path: PathT,
        *,
        class_name: str,
        comments_are: Optional[Union[YAML_COMMENTS_POSITION, Sequence[YAML_COMMENTS_POSITION]]] = "side",
        **kwargs: Any,
    ) -> ObjectT:
        return super().from_yaml_file(loader=loader, path=path, class_name=class_name, comments_are=comments_are)  # type: ignore

    @override
    def from_data(
        self,
        data: ObjectT,
        class_name: str,
        **kwargs: Any,
    ) -> ObjectT:
        if not is_string_python_keyword_compatible(class_name):
            raise LazyTypeHintLiveError(
                f"Given class_name is not compatible with Python class naming conventions: {class_name}"
            )
        custom_class_file_path = self._custom_class_dir_path / f"{class_name}.py"

        if class_name in self._get_classes_added() and self.if_type_hint_exists == "validate":
            string_representation = str(super().from_data(data=data, class_name=class_name))
            if self._remove_docstrings(string_representation) != self._remove_docstrings(
                custom_class_file_path.read_text(encoding="utf-8")
            ):
                raise LazyTypeHintLiveError(
                    "The given object did not generate the same string representation (type hints) for "
                    f"the existing `{class_name}`"
                )
        else:
            string_representation = str(super().from_data(data=data, class_name=class_name))

        self._create_custom_class_py(string_representation, class_name)
        if class_name in self._get_classes_added():
            return data
        self._add_new_class_to_loader_pyi(new_class=class_name)
        return data

    @staticmethod
    def _remove_docstrings(string: str) -> str:
        if '"""' not in string:
            return string

        strings = string.splitlines()
        for idx, string in enumerate(strings):
            strings[idx] = "" if '"""' in string else string
        return "\n".join(strings)

    @classmethod
    def reset(cls) -> None:
        """Remove all existing type hints generated by this class."""
        if cls._custom_class_dir_path.exists():
            shutil.rmtree(cls._custom_class_dir_path)
        if cls._this_file_pyi_path.exists():
            os.remove(cls._this_file_pyi_path)

    @final
    def _generate_this_file_pyi(self) -> PyFileModifier:
        content = Path(__file__).read_text()

        file_handler = PyFileModifier(content)
        file_handler.remove_all_method_bodies()
        file_handler.remove_all_private_methods()
        file_handler.remove_all_instance_variables(class_name=(type(self)).__name__)
        return file_handler

    def _create_custom_class_py(self, string: str, class_name: str) -> None:
        if not self._custom_class_dir_path.exists():
            os.makedirs(self._custom_class_dir_path)

        path = self._custom_class_dir_path / f"{class_name}.py"
        path.write_text(string)

    @final
    def _add_new_class_to_loader_pyi(self, *, new_class: str) -> None:
        self._add_class_created_to_this_file_pyi(new_class)
        self._add_import_to_this_file_pyi(new_class)
        for method_name in self._methods_to_be_overloaded:
            self._add_overload_to_this_file_pyi(new_class=new_class, method_name=method_name)

    @final
    def _add_import_to_this_file_pyi(self, new_class: str) -> None:
        import_statement = f"from lazy_type_hint.generators.build.{new_class} import {new_class}"
        self.this_file_pyi.add_imports(
            import_statement,
            in_type_checking_block=False,
        )
        self._update_this_file_pyi()

    @final
    def _add_overload_to_this_file_pyi(
        self, *, new_class: str, method_name: str, input_argument: str = "class_name"
    ) -> None:
        # First time the function is called it will attach an extra @overload decorator
        if not self.this_file_pyi.search_decorator(decorator_name="overload", method_name=method_name):
            idx_lst = self.this_file_pyi.search_method(method_name, return_index_above_decorator=True)
            if not idx_lst:
                raise LazyTypeHintLiveError(f"No method `{method_name}` could be found")

            self.this_file_pyi.add_line(idx_lst[-1], f"{TAB}@overload")

        signature, _ = self.this_file_pyi.get_signature(method_name)

        # At this point idx is the index of the line where the input argument was found
        first_idx = signature.find(input_argument)
        if first_idx == -1:
            raise LazyTypeHintLiveError(f"No {input_argument} could be found within the signature {signature}")
        first_idx += len(input_argument)
        last_idx = first_idx
        while signature[last_idx] not in (",", ")"):
            last_idx += 1
        # The type hint of the input argument to modify is between first_idx and last_idx
        signature = signature[: first_idx + 1] + f' Literal["{new_class}"]' + signature[last_idx:]

        # Modifying returned value
        last_idx = signature.rfind(":")
        first_idx = signature.rfind("->")
        signature = signature[: first_idx + len("->") + 1] + new_class + signature[last_idx:]

        idx = self.this_file_pyi.search_method(method_name=method_name, return_index_above_decorator=True)[-1]
        self.this_file_pyi.add_line(idx, signature)
        self._update_this_file_pyi()

    @final
    def _add_class_created_to_this_file_pyi(self, new_class: str) -> None:
        label = "classes_created"

        value = self.this_file_pyi.search_assignment(label, only_values=True)[0]
        if not value:
            raise LazyTypeHintLiveError(f"No `{label}` was found in this file.")
        # Case: Any
        if value == "Any":
            value = f'Union[Literal["{new_class}"], Any]'
        # Case: Union[Literal["A"], Any]
        else:
            value = value.replace("], Any]", f', "{new_class}"], Any]')

        self.this_file_pyi.replace_assignement(label, value)
        self._update_this_file_pyi()

    @final
    def _update_this_file_pyi(self) -> None:
        self._this_file_pyi_path.write_text(str(self.this_file_pyi))



from contextlib import suppress

from lazy_type_hint.data_type_tree.generic_type.dict_data_type_tree import DictDataTypeTree as DictDataTypeTree
from lazy_type_hint.data_type_tree.generic_type.list_data_type_tree import ListDataTypeTree as ListDataTypeTree
from lazy_type_hint.data_type_tree.generic_type.mapping_data_type_tree import (
    MappingDataTypeTree as MappingDataTypeTree,
)

with suppress(ImportError):
    from lazy_type_hint.data_type_tree.generic_type.pandas_data_frame_data_type_tree import (
        PandasDataFrameDataTypeTree as PandasDataFrameDataTypeTree,
    )
from lazy_type_hint.data_type_tree.generic_type.iterator_data_type_tree import (
    IteratorDataTypeTree as IteratorDataTypeTree,
)
from lazy_type_hint.data_type_tree.generic_type.set_data_type_tree import SetDataTypeTree as SetDataTypeTree
from lazy_type_hint.data_type_tree.generic_type.tuple_data_type_tree import (
    TupleDataTypeTree as TupleDataTypeTree,
)



import re
from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Final, List, Literal, Mapping, Optional, Sequence, Set, Tuple

from typing_extensions import TypeGuard

from lazy_type_hint.utils.utils import TAB

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias

KEYWORDS_AVAILABLE: "TypeAlias" = Literal[
    "list",
    "set",
    "tuple",
    "dict",
    "type",
    "Sequence",
    "FrozenSet",
    "Hashable",
    "Mapping",
    "TypeAlias",
    "MappingProxyType",
    "TypedDict",
    "Any",
    "Union",
    "Optional",
    "NotRequired",
    "ReadOnly",
    "Callable",
    "Protocol",
    "Iterator",
    "Literal",
    "overload",
    "pandas",
    "pd.Scalar",
    "npt",
    "numpy",
    "ModuleType",
    "TextIO",
]


@dataclass(frozen=True)
class ImportManager:
    _set: Set[KEYWORDS_AVAILABLE] = field(default_factory=set)

    TEMPLATE: Final[str] = field(init=False, default="from {package} import {name}")
    PACKAGE: Mapping[KEYWORDS_AVAILABLE, Tuple[str, ...]] = field(
        default_factory=lambda: {
            "list": ("typing", "List"),
            "set": ("typing", "Set"),
            "tuple": ("typing", "Tuple"),
            "type": ("typing", "Type"),
            "dict": ("typing", "Dict"),
            "Sequence": ("typing", "Sequence"),
            "Hashable": ("typing", "Hashable"),
            "FrozenSet": ("typing", "FrozenSet"),
            "Mapping": ("typing", "Mapping"),
            "MappingProxyType": ("types", "MappingProxyType"),
            "TypedDict": ("typing", "TypedDict"),
            "Any": ("typing", "Any"),
            "Union": ("typing", "Union"),
            "Optional": ("typing", "Optional"),
            "NotRequired": ("typing_extensions", "NotRequired"),
            "ReadOnly": ("typing_extensions", "ReadOnly"),
            "Callable": ("typing", "Callable"),
            "ModuleType": ("types", "ModuleType"),
            "Protocol": ("typing", "Protocol"),
            "Literal": ("typing", "Literal"),
            "overload": ("typing", "overload"),
            "pandas": ("pandas", "pandas"),
            "npt": ("npt", "npt"),
            "numpy": ("numpy", "numpy"),
            "TypeAlias": ("typing_extenions", "TypeAlias"),
            "pd.Scalar": ("pandas._typing", "Scalar"),
            "TextIO": ("typing", "TextIO"),
            "Iterator": ("typing", "Iterator"),
        }
    )

    def add(self, keyword: KEYWORDS_AVAILABLE) -> "Self":
        self._set.add(keyword)
        return self

    def import_all_unkown_symbols_from_signature(self, signature: str) -> None:
        """
        Imports all unknown symbols from the given signature.

        Args:
            signature (str): The signature containing type hints.
        """
        hints = re.findall(r":\s*([^\s,]+)", signature)
        for hint in hints:
            hint_ = str(hint).strip().rstrip().replace("):", "")
            self._import_symbols(hint_)

        pattern = r"->\s*([^\s,]+|$)"
        re_output = re.search(pattern, signature)
        return_value: Optional[str] = None
        if re_output:
            return_value = re_output.group(1).rstrip(":")
            self._import_symbols(return_value)

    def _import_symbols(self, symbol: str) -> None:
        """Import all unknown symbols. Including those belonging to generic type hints."""
        symbols = re.findall(r"\b\w+\b", symbol)
        for symbol in symbols:
            symbol = self._cast_symbol(symbol)
            if self._is_importable_symbol(symbol):
                self.add(symbol)

    @staticmethod
    def _cast_symbol(symbol: str) -> str:
        if symbol in ("List", "Tuple", "Set", "Dict"):
            return symbol.lower()
        return symbol

    def _is_importable_symbol(self, symbol: str) -> TypeGuard[KEYWORDS_AVAILABLE]:
        return symbol in self.PACKAGE

    def format(self, *, line_length: int = 90) -> str:
        import_statements: Dict[str, List[str]] = defaultdict(list)
        for keyword in self._set:
            try:
                reference = self.PACKAGE[keyword]
            except KeyError as error:
                raise ValueError(
                    f"The given keyword `{keyword}` could not be located among the available imports: "
                    f"{', '.join(self.PACKAGE.keys())}"
                ) from error
            import_statements[".".join(reference[:-1])].append(reference[-1])

        imports_lst: List[str] = []
        for package, imports in sorted(import_statements.items(), key=lambda x: x[0]):
            imports_lst.append(self._format_single_package(package, imports, line_length=line_length))
        return "\n".join(imports_lst)

    def _format_single_package(self, package: str, imports: Sequence[str], *, line_length: int) -> str:
        if imports[0] == "pandas":
            string = "import pandas as pd"
        elif imports[0] == "npt":
            string = "import numpy.typing as npt"
        elif imports[0] == "numpy":
            string = "import numpy as np"
        else:
            string = self.TEMPLATE.format(package=package, name=", ".join(sorted(imports)))
        if len(string) > line_length:
            name = f",\n{TAB}".join(sorted(imports))
            string = self.TEMPLATE.format(package=package, name=f"(\n{TAB}{name}")
            string += ",\n)"
        return string

    def __contains__(self, element: object) -> bool:
        return element in self._set



import keyword
from collections import defaultdict
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    Hashable,
    Iterable,
    KeysView,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    cast,
)

from typing_extensions import Self, TypeGuard, override

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.generic_type.mapping_data_type_tree import MappingDataTypeTree
from lazy_type_hint.strategies import ParsingStrategies
from lazy_type_hint.utils import TAB, format_string_as_docstring, is_string_python_keyword_compatible

ValueT = TypeVar("ValueT")


@dataclass(frozen=True)
class DictMetadataComparison:
    common_keys: Set[Hashable] = field(default_factory=set)
    non_common_keys: Set[Hashable] = field(default_factory=set)
    value_types: Mapping[Hashable, Set[Type[object]]] = field(default_factory=lambda: defaultdict(set))

    @property
    def percentage_similarity(self) -> int:
        total_keys = len(self.common_keys) + len(self.non_common_keys)
        if total_keys == 0:
            return 0
        return int(100 * (len(self.common_keys) / total_keys))

    @property
    def all_corresponding_value_types_are_same_type(self) -> bool:
        return all(len(value_types_) == 1 for value_types_ in self.value_types.values())


@dataclass
class KeyInfo:
    name: Hashable
    docstring: str = ""
    required: bool = True


class DictMetadata:
    """Represents the metadata of a dictionary data type."""

    hidden_key_prefix: str
    """Prefix prepended to a key to indicate this one should be hidden when building its type alias."""
    key_info: Dict[Hashable, KeyInfo]
    """Extra information associated to every key of the dictionary."""

    _data: Dict[Hashable, object]
    _strategies: ParsingStrategies
    _initial_keys: KeysView[Hashable]

    def __init__(
        self, data: Mapping[Hashable, object], *, hidden_key_prefix: str, strategies: ParsingStrategies
    ) -> None:
        self._data = dict(data)
        self._initial_keys = data.keys()
        self._strategies = strategies
        self.hidden_key_prefix = hidden_key_prefix
        self.key_info = {}
        self._update_key_info(force_all_required_to_true=True)

    def get_keys(self, *, include_hidden_prefix_keys: bool = False) -> Set[Hashable]:
        """Get all keys within the wrapped dictionary."""
        if include_hidden_prefix_keys:
            return set(self._data.keys())
        if self._all_keys_are_string(self._data):
            return {key for key in self._data if not key.startswith(self.hidden_key_prefix)}
        return set(self._data.keys())

    @staticmethod
    def compare_multiple_dicts(*dicts_metadata: "DictMetadata") -> DictMetadataComparison:
        """Compute different metrics between multiple metadata objects."""
        keys_per_dictionary = [set(object_.get_keys()) for object_ in dicts_metadata]
        if not keys_per_dictionary:
            return DictMetadataComparison()

        common_keys = set.intersection(*keys_per_dictionary)
        total_keys = set.union(*keys_per_dictionary)
        non_common_keys = total_keys - common_keys

        value_types: Dict[Hashable, Set[Type[object]]] = defaultdict(set)
        for dct in dicts_metadata:
            for key, value in dct._data.items():
                value_type = type(value)
                if value_type == float and int in value_types[key]:
                    value_types[key].remove(int)
                value_types[key].add(type(value))
        return DictMetadataComparison(common_keys=common_keys, non_common_keys=non_common_keys, value_types=value_types)

    @property
    def is_typed_dict(self) -> bool:
        if self._all_keys_are_string(self._data) and self._strategies.dict_strategy == "TypedDict":
            return True
        return False

    @property
    def is_functional_syntax(self) -> bool:
        if self._all_keys_are_string(self._data) and self._all_keys_are_parsable():
            return False
        return True

    @staticmethod
    def _all_keys_are_string(dct: Mapping[Hashable, ValueT]) -> "TypeGuard[Mapping[str, ValueT]]":
        return all(isinstance(key, str) for key in dct)

    def _all_keys_are_parsable(self) -> bool:
        if self._all_keys_are_string(self._data) and all(key not in keyword.kwlist for key in self._data):
            return all(is_string_python_keyword_compatible(key) for key in self._data)
        return False

    def update(self, other: "DictMetadata") -> Mapping[Hashable, object]:
        """Update the wrapped dictionary and its metadata by including a new metadata object."""
        for key, value in other._data.items():
            if key not in self._data:
                self._data[key] = value
        self._update_key_info(other._initial_keys)
        return self._data

    def _update_key_info(
        self, keys_that_were_introduced: Optional[KeysView[Hashable]] = None, force_all_required_to_true: bool = False
    ) -> None:
        """Update the key information within this dictionary.

        If `keys_that_were_introduced` is None, all keys will be assumed as `Required`. However, if this one is given,
        a comparison between the existing keys and the given ones is done in order to save which keys are required
        and which ones not.
        """
        docstrings = self.get_key_docstrings(docstring_keys_start_with=self.hidden_key_prefix)

        for key in self._data:
            if key not in self.key_info:
                if isinstance(key, str) and key.startswith(self.hidden_key_prefix):
                    continue
                if force_all_required_to_true:
                    self.key_info[key] = KeyInfo(name=key, required=True)
                else:
                    self.key_info[key] = KeyInfo(name=key, required=False)
            if key in docstrings:
                self.key_info[key].docstring = docstrings[key]

        if force_all_required_to_true:
            return

        if not keys_that_were_introduced:
            for key_info in self.key_info.values():
                key_info.required = False
            return

        non_required_key: Hashable
        for non_required_key in set.difference(set(self._initial_keys), keys_that_were_introduced):
            if isinstance(non_required_key, str) and non_required_key.startswith(self.hidden_key_prefix):
                continue
            self.key_info[non_required_key].required = False

    def get_key_docstrings(
        self, *, docstring_keys_start_with: str = "", return_formatted_as_docstring: bool = False
    ) -> Mapping[Hashable, str]:
        """Obtain a map that associates each key with its corresponding string.

        All those keys that do not have a docstring will not be present in this dictionary.
        """
        if not docstring_keys_start_with:
            docstring_keys_start_with = self.hidden_key_prefix
        if not self._all_keys_are_string(self._data):
            return {}
        dct: Dict[Hashable, str] = {}
        # Insert key docstrings
        for key in self._data:
            if not key.startswith(docstring_keys_start_with):  # Check key is not docstring based
                doc_key: str = f"{self.hidden_key_prefix}{key}"
                if doc_key in self._data:  # Check if there is key docstring
                    unformatted_docstring = self._data[doc_key]  # type: ignore
                    if not isinstance(unformatted_docstring, str):
                        continue
                    if return_formatted_as_docstring:
                        formated_docstring = (
                            "\n" + format_string_as_docstring(unformatted_docstring, indentation=TAB) + "\n"
                        )
                        dct[key] = formated_docstring
                    else:
                        dct[key] = unformatted_docstring
        return dct


class DictDataTypeTree(MappingDataTypeTree):
    wraps = dict  # type: ignore
    data: Dict[Hashable, object]
    dict_metadata: DictMetadata

    @staticmethod
    @override
    def _validate_name(name: str) -> None:
        """Removed name validation as the TypedDict can type any name."""
        return

    @override
    @property
    def permission_to_be_created_as_type_alias(self) -> bool:
        if self.dict_metadata.is_typed_dict:
            return True
        return super().permission_to_be_created_as_type_alias

    @override
    def __pre_child_instantiation__(self) -> None:
        self.dict_metadata = DictMetadata(
            self.data, hidden_key_prefix=self.hidden_keys_prefix, strategies=self.strategies
        )

    @override
    def _get_str_top_node(self) -> str:
        if self.dict_metadata.is_typed_dict:
            children_str_key = cast(Mapping[str, DataTypeTree], self.children)
            return self._parse_typed_dict(children=children_str_key)
        else:
            return self._parse_dict(self.children)

    def _parse_typed_dict(
        self,
        children: Mapping[str, DataTypeTree],
    ) -> str:
        """
        Define the name of the keys and values for the current dictionary.

        If the name of a child type alias needs to be created, this one will use as a suffix the name
        of the associated key.

        Example:
            - address: AddressList
            - age: AgeInt
        """
        self.imports.add("TypedDict")

        content: Dict[str, str] = {}
        for key, value in children.items():
            if isinstance(value, dict):
                type_value = f"{self.name}{self._to_camel_case(key)}"
                content[key] = type_value
            else:
                if not value.permission_to_be_created_as_type_alias:
                    content[key] = value.get_str_top_node_without_lvalue()
                else:
                    name = self._to_camel_case(value.name)
                    content[key] = name
                if key in self.dict_metadata.key_info:
                    if not self.dict_metadata.key_info[key].required:
                        self.imports.add("NotRequired")
                        content[key] = f"NotRequired[{content[key]}]"
                    if self.strategies.typed_dict_read_only_values:
                        self.imports.add("ReadOnly")
                        content[key] = f"ReadOnly[{content[key]}]"
        return self._build_typed_dict(
            name=self.name,
            content=content,
            functional_syntax=self.dict_metadata.is_functional_syntax,
            key_used_as_class_docstring=self.strategies.key_used_as_doc,
        )

    def _build_typed_dict(
        self,
        name: str,
        content: Mapping[str, str],
        *,
        functional_syntax: bool = False,
        key_used_as_class_docstring: str = "",
    ) -> str:
        """
        Build a typed dictionary based on the given name and content.

        Args:
            name (str): The name of the typed dictionary.
            content (Mapping[str, str]): The content of the typed dictionary, where each
                key-value pair represents a field and its type.
            functional_syntax (bool, optional): If True, use functional syntax to define
                the typed dictionary. Defaults to False.
            key_used_as_class_docstring (str, optional): The key to be used as the class docstring.
                This is a hidden one and it will not be represented in the final `TypedDict`.
                Defaults to an empty string.

        Returns:
            str: The string representation of the typed dictionary.
        """
        if functional_syntax:
            template = f"""{name} = TypedDict(
{TAB}"{name}",
{TAB}{{
{TAB}{TAB}"{{key}}": {{value}},
{TAB}}},
)"""
            idx_to_repeat = -3
        else:
            template = f"""class {name}(TypedDict):
{TAB}{{key}}: {{value}}{{optional_key_docstring}}"""
            idx_to_repeat = -1

        # Build the dictionary
        lines = template.splitlines()
        modified_line = ""
        key_docstrings = self.dict_metadata.get_key_docstrings(
            docstring_keys_start_with=self.hidden_keys_prefix, return_formatted_as_docstring=True
        )
        for key, value in content.items():
            if key.startswith(self.hidden_keys_prefix):  # Do not add artificially created keys
                continue
            if self.dict_metadata.is_functional_syntax:
                modified_line += lines[idx_to_repeat].format(key=key, value=value) + "\n"
            else:
                docstring = key_docstrings.get(key, "")
                modified_line += lines[idx_to_repeat].format(key=key, value=value, optional_key_docstring=docstring)
                if not docstring:
                    modified_line += "\n"
        lines[idx_to_repeat] = modified_line[:-1]

        # Append class docstring if found
        if key_used_as_class_docstring in self.data and key_used_as_class_docstring:
            lines = self._insert_class_docstring(lines, key_used_as_doc=key_used_as_class_docstring)
        return "\n".join(lines)

    def update_data_and_metadata(self, other: "DictDataTypeTree") -> None:
        """Given another child, this will update the current node with all the data and metadata."""
        self.data = dict(self.dict_metadata.update(other.dict_metadata))

    def _insert_class_docstring(self, lines: Sequence[str], *, key_used_as_doc: str) -> List[str]:
        string = self.data[key_used_as_doc]
        lines = list(lines)
        if isinstance(string, str):
            indentation = "" if self.dict_metadata.is_functional_syntax else TAB
            docstring = format_string_as_docstring(string, indentation=indentation)
            if self.dict_metadata.is_functional_syntax:
                lines.insert(len(lines), docstring)
            else:
                lines.insert(1, docstring + "\n")
        return lines

    @override
    def _get_hash(self) -> Hashable:
        if not self.dict_metadata.is_typed_dict:
            return super()._get_hash()
        hashes: List[object] = []
        for name, child in self.children.items():
            hashes.append(("typed_dict", name, child._get_hash()))
        return frozenset(hashes)

    @staticmethod
    def compare_multiple_typed_dicts_based_trees(*trees: "DictDataTypeTree") -> DictMetadataComparison:
        """
        Obtain metrics related to the comparison of multiple objects.

        This allows to check its similarity or keys in common for example.
        """
        dicts_metadata = [
            tree.dict_metadata
            for tree in trees
            if isinstance(tree, DictDataTypeTree) and tree.dict_metadata.is_typed_dict
        ]
        return DictMetadata.compare_multiple_dicts(*dicts_metadata)

    @classmethod
    def from_multiple_dict_data_type_trees(cls, *trees: "DictDataTypeTree") -> "Self":
        """
        Merge multiple dict data type trees.

        For thsis to happen, two main tasks are carried out:
            - Update `data` that holds the dictionary
            - Override `DictMetadata` with new information relative to all new information
        """
        merged_dict: Dict[Hashable, object] = {}
        for d in (tree.data for tree in trees):
            merged_dict.update(d)

        def check_same(trees: Iterable["DictDataTypeTree"], *, param: str) -> Any:
            if not trees:
                raise ValueError("At least one tree must be given")

            first_tree = next(iter(trees))
            if all(getattr(first_tree, param) == getattr(tree, param) for tree in trees):
                return getattr(first_tree, param)
            raise ValueError(f"Only {type(first_tree).__name__} with same `{param}` can be merged.")

        random_tree = next(iter(trees))
        shortest_name = ""
        for tree in trees:
            random_tree.dict_metadata.update(tree.dict_metadata)
            if not shortest_name:
                shortest_name = tree.name
            shortest_name = min(shortest_name, tree.name)

        new_tree = cls(
            data=merged_dict,
            name=shortest_name,
            imports=check_same(trees, param="imports"),
            depth=check_same(trees, param="depth"),
            strategies=check_same(trees, param="strategies"),
            parent=trees[0].parent,
        )
        new_tree.dict_metadata = random_tree.dict_metadata
        return new_tree



from typing import Any, Iterator, Sequence, Tuple

from typing_extensions import override

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.generic_type.sequence_data_type_tree import (
    SequenceDataTypeTree,
)


class IteratorDataTypeTree(SequenceDataTypeTree):
    wraps = Iterator

    @override
    def _instantiate_children(self, data: Sequence[Any]) -> Tuple[DataTypeTree, ...]:  # type: ignore
        return self.operations.instantiate_children(data, allow_repeated_children=False)

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("Iterator")
        container = "Iterator"
        return f"{self.name} = {container}[{self.get_type_alias_children()}]"



import subprocess
from pathlib import Path
from typing import Iterable, List, NamedTuple, Union

from lazy_type_hint.utils.utils import check_if_command_available


class Mypy:
    """
    A class for running the Mypy static type checker.

    Attributes:
        Result: A named tuple representing the result of running Mypy.
    """

    @staticmethod
    def is_available() -> bool:
        return check_if_command_available("python3 -m poetry run mypy")

    class Result(NamedTuple):
        """Results returned by Mypy."""

        success: bool
        errors: List[str]
        scanned: Union[Path, str]

        def errors_as_str(self) -> str:
            return "\n".join(self.errors)

        def file_content_with_lineno(self) -> str:
            if isinstance(self.scanned, str):
                raise ValueError("This method is only callable when a file has been scanned.")
            lines: List[str] = []
            for idx, line in enumerate(self.scanned.read_text().splitlines()):
                lines.append(f"{idx} {line}")
            return "\n".join(lines)

        def __str__(self) -> str:
            if self.success:
                return "No errors"
            else:
                string = "Errors were found"
            if isinstance(self.scanned, Path):
                return (
                    f"{string}. {self.errors_as_str()}\n\nFile scanned ({self.scanned.name}): "
                    f"\n{self.file_content_with_lineno()}"
                )
            return f"{string}. {self.errors_as_str()}"

    def scan_string(
        self, string: str, *, strict: bool = False, python_version: str = "3.8", ignore_errors: Iterable[str] = ()
    ) -> Result:
        """
        Scan a string using the mypy static type checker.

        Args:
            string (str): The string to be scanned.
            strict (bool, optional): Whether to enable strict mode. Defaults to False.
            python_version (str, optional): The Python version to use for type checking. Defaults to "3.8".
            ignore_errors (Iterable[str], optional): A collection of error keywords to ignore during type checking.
                Defaults to ().

        Returns:
            Result: The result of the type checking, including success status, errors, and the scanned string.
        """
        command = ["python3", "-m", "poetry", "run", "mypy"]
        if strict:
            command.append("--strict")
        command.append("--python-version")
        command.append(python_version)
        command.append("-c")
        command.append(string)
        result = subprocess.run(command, capture_output=True)
        if result.stderr:
            return self.Result(success=False, errors=[str(result.stderr)], scanned=string)
        stdout = str(result.stdout)
        errors: List[str] = []
        lines = stdout.split("error: ")
        for error in lines:
            error_str = error[: error.rfind("]") + 1]
            if error_str:
                if ignore_errors:
                    if all(keyword not in error_str for keyword in ignore_errors):
                        errors.append(error_str)
                else:
                    errors.append(error_str)
        return self.Result(success=not bool(errors), errors=errors, scanned=string)

    def scan_file(
        self,
        file: Union[str, Path],
        *,
        strict: bool = False,
        python_version: str = "3.8",
        ignore_errors: Iterable[str] = (),
    ) -> Result:
        """
        Run Mypy on the specified file.

        Args:
            file (Union[str, Path]): The file to run Mypy on.
            strict (bool, optional): Whether to enable strict mode. Defaults to False.
            python_version (str, optional): The Python version to use. Defaults to "3.8".
            ignore_errors (Iterable[str], optional): A collection of error strings to
                ignore. Defaults to ().

        Returns:
            Result: The result of running Mypy.

        """
        command = ["mypy"]
        if strict:
            command.append("--strict")
        command.append(f"--python-version {python_version}")
        command.append(str(file))
        result = subprocess.run(" ".join(command), capture_output=True, text=True)

        if result.stderr:
            raise ValueError(f"Mypy could not run: {result.stderr}")

        error_lst = [line for line in result.stdout.splitlines() if "error: " in line]
        if ignore_errors:
            error_lst = [line for line in error_lst if not all(ignore_error in line for ignore_error in ignore_errors)]
        # Filter out to only get the errors from the file
        error_lst = [error for error in error_lst if Path(file).name in error]
        return self.Result(
            success=not bool(error_lst),
            errors=error_lst,
            scanned=Path(file),
        )



import ast
from collections.abc import Sequence
from typing import Any, List, Literal, Tuple, Union, overload

from lazy_type_hint.utils import TAB


class PyFileModifier:
    lines: List[str]

    def __init__(self, representation: str) -> None:
        self.lines = representation.splitlines()

    def __str__(self) -> str:
        return "\n".join(self.lines)

    def __repr__(self) -> str:
        return "\n".join(self.lines)

    @overload
    def search_assignment(self, variable: str, only_values: Literal[False] = False) -> List[Tuple[int, str]]:
        ...

    @overload
    def search_assignment(self, variable: str, only_values: Literal[True]) -> List[str]:
        ...

    def search_assignment(self, variable: str, only_values: bool = False) -> Union[List[Tuple[int, str]], List[str]]:
        """
        Searches for assignments of a given variable in the lines of the file.

        Args:
            variable (str): The variable to search for assignments.
            only_values (bool): Only return the values being assigned to that variable.

        Returns:
            List[Tuple[int, str]]: A list of tuples containing the line index and the
                assigned value.
        """
        lst: List[Any] = []
        for idx, line in enumerate(self.lines):
            if variable in line and "=" in line and not self.it_is_string(line, variable):
                value = line.split("=")[-1].strip()
                if isinstance(value, str) and value[0] in ("'", '"'):
                    value = value[1:-1]
                if only_values:
                    lst.append(value)
                else:
                    lst.append((idx, value))
        return lst

    def search_decorator(self, *, decorator_name: str, method_name: str) -> List[int]:
        """
        Searches for a specific decorator in the lines of code preceding a method.

        Args:
            decorator_name (str): The name of the decorator to search for.
            method_name (str): The name of the method to search for.

        Returns:
            List[int]: A list of line numbers where the decorator is found.
        """
        lst = self.search_method(method_name=method_name, return_index_above_decorator=False)
        output: List[int] = []
        for idx in lst:
            while "@" in self.lines[idx - 1]:
                if f"@{decorator_name}" in self.lines[idx - 1]:
                    output.append(idx - 1)
                    break
                break
        return output

    def search_method(self, method_name: str, *, return_index_above_decorator: bool = True) -> List[int]:
        """
        Search for a method in the lines of the file and return its location (indices).

        Args:
            method_name (str): The name of the method to search for.
            return_index_above_decorator (bool, optional): Whether to return the index
                above the decorator if found. Defaults to True.

        Returns:
            List[int]: A list of indices where the method is found.
        """
        lst: List[int] = []
        for idx, line in enumerate(self.lines):
            if line.strip().startswith(f"def {method_name}") and not self.it_is_string(line, method_name):
                if return_index_above_decorator:
                    while "@" in self.lines[idx - 1]:
                        idx = idx - 1
                lst.append(idx)
        return lst

    def it_is_string(self, line: str, keyword_of_interest: str) -> bool:
        for string_symbol in ('"', "'"):  # noqa: SIM110
            if string_symbol in line.split(keyword_of_interest)[0]:
                return True
        return False

    def add_line(self, idx: int, line: Union[str, Sequence[str]]) -> None:
        """
        Add a line or multiple lines at the specified index.

        Args:
            idx (int): The index at which to insert the line(s).
            line (Union[str, Sequence[str]]): The line(s) to be added. It can be a
                single string or a sequence of strings.
        """
        if not isinstance(line, str):
            line = "\n".join(line)
        self.lines.insert(idx, line)

    def search_line(self, keyword: str) -> int:
        """
        Searches for a line containing the specified keyword within the file.

        Args:
            keyword (str): The keyword to search for.

        Returns:
            int: The index of the line containing the keyword.
        """
        for idx, line in enumerate(self.lines):
            if keyword in line:
                return idx
        raise ValueError(f"Line with `{keyword}` could not be found within the file.")

    def replace_line(self, idx: int, line: str) -> None:
        """
        Replaces the line at the specified index with the given line.

        Args:
            idx (int): The index of the line to be replaced.
            line (str): The new line to replace the existing line.
        """
        self.lines[idx] = line

    def replace_assignement(self, label: Union[int, str], value: str) -> None:
        """
        Replaces the value of an assignment line identified by the given label or index.

        Args:
            label (Union[int, str]): The label or index of the assignment line to be
                replaced.
            value (str): The new value to be assigned.
        """
        if not isinstance(label, int):
            assignment_idx = self.search_assignment(label)
            if not assignment_idx:
                raise ValueError(f"The given label ({label}) is not being assigned.")
            if len(assignment_idx) > 1:
                raise ValueError(
                    f"The given label ({label}) is assigned multiple times. Not clear which one should be replaced."
                )
            idx = assignment_idx[0][0]
        else:
            idx = label

        line = self.lines[idx]
        if "=" not in line:
            raise ValueError(f"The given line ({idx}) is not an assignment one.")
        symbols = line.split("=")
        symbols[-1] = value
        self.lines[idx] = "=".join(symbols)

    def remove_all_method_bodies(self) -> None:
        """
        Removes the bodies of all methods in the file.

        This method parses the file's contents using the `ast` module and walks through
        the abstract syntax tree (AST) to find all function definitions. For each function
        definition found, the body of the function is replaced with an `Ellipsis` object,
        effectively removing the method body.

        After modifying the AST, the updated code is converted back to a string and split
        into lines, which are then stored in the `lines` attribute of the object.

        Note: This method modifies the object in-place.
        """
        tree = ast.parse(str(self))
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                node.body = [ast.Ellipsis()]  # type: ignore
        self.lines = ast.unparse(tree).strip().splitlines()  # type: ignore

    def remove_all_private_methods(self) -> None:
        """
        Removes all private methods from the file.

        This method iterates through the lines of the file and removes any lines
        that define a private method. A private method is defined as a method that
        starts with an underscore (_) character.

        Note: This method modifies the object in-place.
        """
        lines_to_remove: List[int] = []
        for idx, line in enumerate(self.lines):
            line_ = line.strip()
            if line_.startswith("def _") and not line_.startswith("def __"):
                idx_ = idx - 1
                # Remove all decorators
                if "@" in self.lines[idx_]:
                    while "@" in self.lines[idx_]:
                        lines_to_remove.append(idx_)
                        idx_ -= 1

                # Remove main function
                lines_to_remove.append(idx)

                idx_ = idx
                # Remove all arguments if they are in multiple lines
                while "):" not in self.lines[idx_] and "->" not in self.lines[idx_]:
                    idx_ += 1
                    lines_to_remove.append(idx_)

                # Remove body of the method
                # Count tabs
                spaces = 0
                for char in line:
                    if char.isspace():
                        spaces += 1
                    else:
                        break

                idx_ += 1
                while len(self.lines) > idx_ and len(self.lines[idx_]) > spaces and self.lines[idx_][spaces].isspace():
                    lines_to_remove.append(idx_)
                    idx_ += 1
                if len(self.lines) > idx_:
                    lines_to_remove.append(idx_)

        for idx in sorted(lines_to_remove)[::-1]:
            self.lines.pop(idx)

    def remove_all_instance_variables(self, class_name: str) -> None:
        """
        Removes all instance variables from a class.

        Args:
            class_name (str): The name of the class.
        """
        upper_idx = self.search_line(f"class {class_name}(") + 1

        lower_idx = self.search_method("__init__", return_index_above_decorator=True)[-1]
        for idx in list(range(upper_idx, lower_idx))[::-1]:
            if "classes_created" in self.lines[idx] or "classes_created" in self.lines[idx - 1]:
                continue
            self.lines.pop(idx)

    def add_imports(self, lines: Union[str, Sequence[str]], *, in_type_checking_block: bool = False) -> None:
        """
        Adds import statements to the file.

        Args:
            lines (Union[str, Sequence[str]]): The import statements to add. It can
                be a single string or a sequence of strings.
            in_type_checking_block (bool, optional): Specifies whether the import
                statements should be added inside a type checking block.
                Defaults to False.
        """
        idx = 0
        if isinstance(lines, str):
            lines = [lines]

        if in_type_checking_block:
            idx = self.search_line("if TYPE_CHECKING:") + 1
            lines = [f"{TAB}{line}" for line in lines]
        for line in lines[::-1]:
            self.lines.insert(idx, line)

    def get_signature(self, method_name: str) -> Tuple[str, slice]:
        """
        Retrieves the signature of a method.

        Args:
            method_name (str): The name of the method.

        Returns:
            Tuple[str, slice]: A tuple containing the method signature as a string and a
                slice object representing the range of lines in the file that contain
                the signature.
        """
        upper_idx = self.search_method(method_name, return_index_above_decorator=True)[-1]
        bottom_idx = upper_idx

        while ")" not in self.lines[bottom_idx] and ":" not in self.lines[bottom_idx]:
            bottom_idx += 1

        return "\n".join(self.lines[upper_idx : bottom_idx + 1]), slice(upper_idx, bottom_idx + 1)



from types import MappingProxyType

from lazy_type_hint.data_type_tree.generic_type.mapping_data_type_tree import MappingDataTypeTree


class MappingProxyDataTypeTree(MappingDataTypeTree):
    wraps = MappingProxyType



import re
from types import MappingProxyType
from typing import Dict, Final, Hashable, Iterator, List, Literal, Mapping, Set

from typing_extensions import override

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.factory import data_type_tree_factory
from lazy_type_hint.data_type_tree.generic_type.generic_data_type_tree import GenericDataTypeTree
from lazy_type_hint.file_modifiers.yaml_file_modifier import YamlFileModifier


class MappingDataTypeTree(GenericDataTypeTree):
    children: Mapping[Hashable, DataTypeTree]
    wraps = MappingProxyType
    hidden_keys_prefix: Final = YamlFileModifier.prefix

    # Iterable-protocol related
    _keys: Iterator[Hashable]

    @override
    def _instantiate_children(self, data: Mapping[Hashable, object]) -> Mapping[Hashable, DataTypeTree]:  # type: ignore
        children: Dict[Hashable, DataTypeTree] = {}
        for key, value in data.items():
            suffix = type(key).__name__ if not isinstance(key, str) else self._to_camel_case(key)
            if isinstance(key, str) and key.startswith(self.hidden_keys_prefix):
                continue
            child = data_type_tree_factory(
                data=value,
                name=f"{self.name}{suffix}",
                imports=self.imports,
                depth=self.depth + 1,
                strategies=self.strategies,
                parent=self,
            )
            if child in children.values():
                self._replace_old_childs_by_new_one(child, children)
                children[key] = child
            else:
                children[key] = child
        return children

    def _replace_old_childs_by_new_one(self, child: DataTypeTree, children: Dict[Hashable, DataTypeTree]) -> None:
        """
        Replaces old child nodes with a new one in the given dictionary of children.

        Only replaced those children that are equal to the given child. In addition, naming is updated to
        make them match.
        """
        name = f"{self.name}{type(child.data).__name__.capitalize()}"
        modified_name = name
        count = 2
        while modified_name in children.values():
            modified_name = f"{name}{count}"
            count += 1
        child.name = modified_name
        for inserted_key in children:
            if children[inserted_key] == child:
                children[inserted_key] = child

    def _get_str_top_node(self) -> str:
        return self._parse_dict(self.children)

    def _parse_dict(self, children: Mapping[Hashable, DataTypeTree]) -> str:
        """Get a string representation of the dictionary for this same top node.

        Examples:
            - Mapping[str, Union[float, str]]
            - Mapping[str, str]
        """
        container: Literal["dict", "TypedDict", "MappingProxyType", "Mapping"] = (
            "dict" if self.strategies.dict_strategy == "TypedDict" else self.strategies.dict_strategy
        )
        if self.holding_type.__name__ == "mappingproxy":
            container = "MappingProxyType"
            self.imports.add(container)
        else:
            self.imports.add(container)

        keys = self._get_types(iterable=children.keys(), remove_repeated=True)
        keys_str = self._format_types(keys)
        value_types = self.get_type_alias_children()

        container_ = "Dict" if container == "dict" else container
        return f"{self.name} = {container_}[{keys_str}, {value_types}]"

    @staticmethod
    def _to_camel_case(string: str) -> str:
        """Make it camel case and compatible with Python keywords."""
        # Remove any initial number within the string
        match = re.match(r"^\d+", string)
        if match:
            number = match.group()
            string = string[len(number) :]

        new_string = [""] * len(string)
        idx_to_remove: Set[int] = set()
        removed = False
        for idx, char in enumerate(string):
            if not char.isalpha() and not char.isnumeric():
                idx_to_remove.add(idx)
            if idx not in idx_to_remove:
                if removed is True:
                    new_string[idx] = string[idx].upper()
                    removed = False
                else:
                    new_string[idx] = string[idx]
            else:
                removed = True

        new_string[0] = new_string[0].upper()
        return "".join(new_string)

    @override
    def _get_hash(self) -> Hashable:
        hashes: List[object] = []
        for name, child in self.children.items():
            hashes.append(("mapping", hash(type(name)), child._get_hash()))
        return frozenset(hashes)



from typing import Hashable, List

from typing_extensions import override

from lazy_type_hint.data_type_tree.generic_type.generic_data_type_tree import (
    GenericDataTypeTree,
)
from lazy_type_hint.data_type_tree.generic_type.set_and_sequence_operations import SetAndSequenceOperations


class SequenceDataTypeTree(GenericDataTypeTree):
    operations: SetAndSequenceOperations

    @override
    def __pre_child_instantiation__(self) -> None:
        self.operations = SetAndSequenceOperations(self)

    @override
    def _get_hash(self) -> Hashable:
        hashes: List[object] = []
        for child in self:
            hashes.append(child._get_hash())
        return frozenset(hashes)



import builtins

from typing_extensions import override

from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree


class InstanceDataTypeTree(SimpleDataTypeTree):
    # Change it by `NoneType` once I drop support with Python 3.8
    wraps = (bool, int, float, range, slice, str, type(None))  # + Custom classes

    @override
    def _get_str_top_node(self) -> str:
        if self.holding_type == type(None):
            self.imports.add("Optional")
            return f"{self.name} = Optional[object]"

        if self._is_builtin_class():
            return f"{self.name} = {self.holding_type.__name__}"
        return f'{self.name} = "{self.holding_type.__name__}"'

    @override
    def _check_tree_is_correct_one(self, data: object) -> None:
        return

    def _is_builtin_class(self) -> bool:
        try:
            cls = getattr(builtins, self.holding_type.__name__)
            return isinstance(cls, type)
        except AttributeError:
            return False



from dataclasses import dataclass, fields
from typing import Literal, get_args, get_type_hints

LIST_STRATEGIES = Literal["Sequence", "list"]
TUPLE_SIZE_STRATEGIES = Literal["fixed", "any size"]
MAPPING_STRATEGIES = Literal["TypedDict", "Mapping", "dict"]
PANDAS_STRATEGIES = Literal["Full type hint", "Type hint only for autocomplete", "Do not type hint columns"]


@dataclass(frozen=True)
class ParsingStrategies:
    list_strategy: LIST_STRATEGIES = "list"
    tuple_size_strategy: TUPLE_SIZE_STRATEGIES = "fixed"
    dict_strategy: MAPPING_STRATEGIES = "TypedDict"
    pandas_strategies: PANDAS_STRATEGIES = "Full type hint"
    min_height_to_define_type_alias: int = 1
    key_used_as_doc: str = ""
    merge_different_typed_dicts_if_similarity_above: int = 50
    typed_dict_read_only_values: bool = False

    def __post_init__(self) -> None:
        type_hints = get_type_hints(self)
        for field in fields(type(self)):
            allowed_values = get_args(type_hints[field.name])
            if allowed_values and getattr(self, field.name) not in allowed_values:
                raise ValueError(
                    f"Invalid value for {field.name}. Expected any of ({', '.join(allowed_values)}) but got "
                    f"{getattr(self, field.name)}"
                )
        if self.min_height_to_define_type_alias < 0:
            raise ValueError("`min_height_to_define_type_alias` must be greater or equal than 0")

        if self.merge_different_typed_dicts_if_similarity_above <= 0:
            raise ValueError("`merge_typed_dicts_if_similarity_above` must be greater than 0")
        if self.merge_different_typed_dicts_if_similarity_above > 100:
            raise ValueError("`merge_typed_dicts_if_similarity_above` must be less than 100")



import builtins
from typing import Type

from typing_extensions import override

from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree


class TypeDataTypeTree(SimpleDataTypeTree):
    wraps = type
    data: Type[object]

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("type")
        if self.is_builtin_class():
            return f"{self.name} = Type[{self.data.__name__}]"
        return f'{self.name} = Type["{self.data.__name__}"]'

    def is_builtin_class(self) -> bool:
        try:
            cls = getattr(builtins, self.data.__name__)
            return isinstance(cls, type)
        except AttributeError:
            return False



"""This module groups common operations that are applied to both sets and sequences.

Although they present a total different behaviour at runtime, from the point of view of type
hinting they are quite similar.
"""

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from lazy_type_hint.data_type_tree.factory import data_type_tree_factory
from lazy_type_hint.data_type_tree.generic_type.dict_data_type_tree import DictDataTypeTree

if TYPE_CHECKING:
    from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.sequence_data_type_tree import SequenceDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.set_data_type_tree import SetDataTypeTree


@dataclass(frozen=True)
class SetAndSequenceOperations:
    data_type_tree: "Union[SetDataTypeTree, SequenceDataTypeTree]"

    def instantiate_children(self, data: Sequence[Any], *, allow_repeated_children: bool) -> Tuple["DataTypeTree", ...]:
        """Instantiate the children for sets and sequences.

        If `allow_repeated_children` is set to True, all children will be returned even if they are repeated.
        """
        if allow_repeated_children:
            children: Union[Set[DataTypeTree], List[DataTypeTree]] = []
        else:
            children = set()
        names_added: Dict[DataTypeTree, str] = {}  # Used to generate new and unique cnames in a quicker way.

        child: DataTypeTree
        for element in data:
            name = f"{self.data_type_tree.name}{type(element).__name__.capitalize()}"
            # Generate new name in case this one was already added
            if name in names_added.values():
                modified_name = name
                count = 2
                while modified_name in names_added.values():
                    modified_name = f"{name}{count}"
                    count += 1
                name = modified_name
            child = data_type_tree_factory(
                data=element,
                name=name,
                imports=self.data_type_tree.imports,
                depth=self.data_type_tree.depth + 1,
                parent=self.data_type_tree,
                strategies=self.data_type_tree.strategies,
            )
            if allow_repeated_children:
                children = cast("List[DataTypeTree]", children)
                if child in names_added:
                    # Rebuild the tree with the name that is already defined
                    child = data_type_tree_factory(
                        data=element,
                        name=names_added[child],
                        imports=self.data_type_tree.imports,
                        depth=self.data_type_tree.depth + 1,
                        parent=self.data_type_tree,
                        strategies=self.data_type_tree.strategies,
                    )
                children.append(child)
                names_added[child] = child.name
            else:
                children = cast("Set[DataTypeTree]", children)
                if child in children and isinstance(child, DictDataTypeTree) and child.dict_metadata.is_typed_dict:
                    self._update_existing_typed_dict_child_from_another_equal_child(children, child)
                if child not in children:
                    children.add(child)
                    names_added[child] = name

        return self._merge_similar_typed_dicts(
            children,
            merge_if_similarity_above=self.data_type_tree.strategies.merge_different_typed_dicts_if_similarity_above,
            allow_repeated_children=allow_repeated_children,
        )

    def _update_existing_typed_dict_child_from_another_equal_child(
        self, children: "Iterable[DataTypeTree]", child: DictDataTypeTree
    ) -> None:
        """
        Merge dicts equal from the point of view of types but that might carry any extra information or metadata.

        This extra information can be hidden docstrings. Therefore, this will transfer all the information
        to the already added DictDataTypeTree child
        """
        for child_added in children:
            if child == child_added and isinstance(child_added, DictDataTypeTree):
                child_added.update_data_and_metadata(child)

    @staticmethod
    def _merge_similar_typed_dicts(
        children: "Union[Set[DataTypeTree], Sequence[DataTypeTree]]",
        *,
        merge_if_similarity_above: int,
        allow_repeated_children: bool,
    ) -> "Tuple[DataTypeTree, ...]":
        """Merge similar TypedDicts.

        Only those TypedDict based children are taken into account. If the similarity is above the expected one,
        a new child that contains a merged dictionary with the merged metadata is created. Then, all those TypedDict
        based children are replaced within `children` by the newly created merged child.
        """
        comparison = DictDataTypeTree.compare_multiple_typed_dicts_based_trees(*children)
        if comparison.percentage_similarity < merge_if_similarity_above:
            return tuple(children)
        if not comparison.all_corresponding_value_types_are_same_type:
            return tuple(children)

        dict_data_type_trees: List[DictDataTypeTree] = []
        if allow_repeated_children:
            children = list(children)
            for child in list(children):
                if isinstance(child, DictDataTypeTree) and child.dict_metadata.is_typed_dict:
                    dict_data_type_trees.append(child)
            merged_child = DictDataTypeTree.from_multiple_dict_data_type_trees(*dict_data_type_trees)
            for idx, child in enumerate(children):
                if isinstance(child, DictDataTypeTree) and child.dict_metadata.is_typed_dict:
                    children[idx] = merged_child
            return tuple(children)
        else:
            children = cast("Set[DataTypeTree]", children)
            for child in children.copy():
                if isinstance(child, DictDataTypeTree) and child.dict_metadata.is_typed_dict:
                    children.remove(child)
                    dict_data_type_trees.append(child)

            merged_child = DictDataTypeTree.from_multiple_dict_data_type_trees(*dict_data_type_trees)
            children.add(merged_child) if isinstance(children, list) else children.add(merged_child)
            return tuple(children)



from lazy_type_hint.utils.docstring_formatter import format_string_as_docstring as format_string_as_docstring
from lazy_type_hint.utils.import_manager import ImportManager as ImportManager
from lazy_type_hint.utils.mypy import Mypy as Mypy
from lazy_type_hint.utils.ordered_set import OrderedSet as OrderedSet
from lazy_type_hint.utils.utils import (
    TAB as TAB,
)
from lazy_type_hint.utils.utils import (
    cache_returned_value as cache_returned_value,
)
from lazy_type_hint.utils.utils import (
    check_if_command_available as check_if_command_available,
)
from lazy_type_hint.utils.utils import (
    is_string_python_keyword_compatible as is_string_python_keyword_compatible,
)



from typing import Hashable, Union, final

from typing_extensions import Self, override

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree


class SimpleDataTypeTree(DataTypeTree):
    """Tree that holds any kind of object that cannot contain inner structures (children)."""

    children: None

    @final
    @override
    def _instantiate_children(self, data: Union[bool, float, str]) -> None:  # type: ignore
        return None

    @final
    @override
    def __iter__(self) -> "Self":
        return self

    @final
    @override
    def __next__(self) -> DataTypeTree:
        raise StopIteration

    @override
    def _get_hash(self) -> Hashable:
        return id(self.holding_type)

    @final
    @override
    def _get_height(self) -> int:
        return 0

    @override
    @property
    def permission_to_be_created_as_type_alias(self) -> bool:
        if self.parent is None:
            return True
        return False



import ast
import os
import re
import subprocess
from itertools import zip_longest
from typing import Any, Final, List, Protocol, TypeVar, Union, cast

TAB: Final = "    "


def is_string_python_keyword_compatible(string: str) -> bool:
    """
    Checks if a string is compatible with Python keywords.

    Args:
        string (str): The string to be checked.

    Returns:
        bool: True if the string is compatible with Python keywords, False otherwise.
    """
    if bool(re.compile(r"^[a-zA-Z0-9_]+$").match(string)):
        if not string[0].isnumeric():
            return True
    return False


def compare_ast(
    node1: Union[ast.expr, List[ast.expr], ast.Module],
    node2: Union[ast.expr, List[ast.expr], ast.Module],
    ignore_args: bool = False,
    ignore_imports: bool = False,
) -> bool:
    """Compare two AST nodes for equality."""
    if type(node1) is not type(node2):
        return False

    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in {"lineno", "end_lineno", "col_offset", "end_col_offset", "ctx"}:
                continue
            if ignore_args and k == "args":
                continue
            if ignore_imports and k == "body" and isinstance(v, list):
                v = [n for n in v if not isinstance(n, (ast.Import, ast.ImportFrom))]
                v2 = getattr(node2, k)
                v2 = [n for n in v2 if not isinstance(n, (ast.Import, ast.ImportFrom))]
                setattr(node2, k, v2)
            if not compare_ast(
                v,
                getattr(node2, k),
                ignore_args=ignore_args,
                ignore_imports=ignore_imports,
            ):
                return False
        return True

    elif isinstance(node1, list) and isinstance(node2, list):
        return all(
            compare_ast(n1, n2, ignore_args=ignore_args, ignore_imports=ignore_imports)
            for n1, n2 in zip_longest(node1, node2)
        )
    else:
        return node1 == node2  # type: ignore


class _AnyMethodProtocol(Protocol):
    __name__: str

    def __call__(*args: Any, **kwargs: Any) -> Any:
        ...


_AnyMethodT = TypeVar("_AnyMethodT", bound=_AnyMethodProtocol)


def cache_returned_value(method: _AnyMethodT) -> _AnyMethodT:
    """
    A decorator that caches the returned value of a method.

    This decorator can be used to cache the returned value of a method
    to improve performance by avoiding redundant computations.

    Args:
        method: The method to be decorated.

    Returns:
        The decorated method.

    Usage:
        @cache_returned_value
        def my_method(self, args, **kwargs):
            # method implementation

    Example:
        class MyClass:
            @cache_returned_value
            def my_method(self, args, **kwargs):
                # method implementation
    """

    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        cache_attr = f"{method.__name__}____"
        if hasattr(self, cache_attr):
            return getattr(self, cache_attr)
        else:
            result = method(self, *args, **kwargs)
            setattr(self, cache_attr, result)
            return result

    return cast(_AnyMethodT, wrapper)


def check_if_command_available(tool: str) -> bool:
    """
    Check if a command is available.

    Args:
        tool (str): The name of the command to check.

    Returns:
        bool: True if the command is available, False otherwise.
    """
    result = subprocess.run(f"{tool} --help", stdout=open(os.devnull, "wb"), stderr=open(os.devnull, "wb"), shell=True)
    return bool(not result.returncode)



from lazy_type_hint.utils.ordered_set import OrderedSet


def test_ordered_set() -> None:
    # Create an instance of OrderedSet
    ordered_set = OrderedSet[int]()

    # Add elements to the ordered set
    ordered_set.add(1)
    ordered_set.add(2)
    ordered_set.add(3)

    # Check the elements as a tuple
    assert ordered_set.as_tuple() == (1, 2, 3)

    # Check the elements as a list
    assert ordered_set.as_list() == [1, 2, 3]

    # Add a duplicate element
    ordered_set.add(2)

    # Check that the duplicate element is not added again
    assert ordered_set.as_tuple() == (1, 2, 3)
    assert ordered_set.as_list() == [1, 2, 3]



"""`DataTypeTree` object that creates a tree representing all container types within the data type given.

This data can then be parsed to a .py compatible file that will type hint your code.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    FrozenSet,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    final,
)

from lazy_type_hint.strategies import ParsingStrategies
from lazy_type_hint.utils import (
    ImportManager,
    OrderedSet,
    cache_returned_value,
    is_string_python_keyword_compatible,
)

if TYPE_CHECKING:
    from typing_extensions import Self


class DataTypeTreeError(Exception):
    ...


DataTypeT = TypeVar("DataTypeT", bound=object)
ChildrenStructure = Union[FrozenSet[DataTypeT], Sequence[DataTypeT], Mapping[Hashable, DataTypeT]]
"""Different child structures that the Tree can hold."""


class DataTypeTree(ABC):
    """Tree that represents any kind of data with its inner structures."""

    name: str
    """Name that represents this node."""
    depth: int
    """Depth of the current node with respect to the whole tree."""
    height: int
    """Maximum height among all the child nodes."""
    data: object
    """Data that was given as input to parse.
    
    This one might have been modified with respect to the original one.
    """
    imports: ImportManager  # Unique one shared among the whole tree
    """Handle the imports required to generate the string representation."""
    parent: Optional[DataTypeTree]
    """Parent node (if any)."""
    children: Optional[ChildrenStructure[DataTypeTree]]
    """All children available within the tree."""
    holding_type: Type[object]
    """Type of input data given."""
    strategies: ParsingStrategies
    """Strategies to follow when parsing the data."""

    subclasses: ClassVar[Mapping[Type[object], Type[DataTypeTree]]] = {}
    """Available subclasses according to the type they are able to parse."""
    wraps: ClassVar[Union[Type[object], Sequence[Type[object]]]] = object
    """Object type that the tree is able to parse."""

    @final
    def __init__(
        self,
        data: object,
        name: str,
        *,
        imports: Optional[ImportManager] = None,
        depth: int = 0,
        strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
        parent: Optional[DataTypeTree] = None,
    ) -> None:
        # Validation
        self._validate_name(name)
        self._check_tree_is_correct_one(data)

        self.data = data
        self.name = name
        self.holding_type = type(data)
        self.strategies = strategies
        self.depth = depth
        self.imports = ImportManager() if imports is None else imports
        self.parent = parent
        self.__pre_child_instantiation__()
        self.children = self._instantiate_children(data)
        self.height = self._get_height()
        self.__post_child_instantiation__()

    @classmethod
    def get_subclass(cls, data: object) -> Type[DataTypeTree]:
        for subclass in DataTypeTree.subclasses.values():
            wraps = [subclass.wraps] if not isinstance(subclass.wraps, Iterable) else subclass.wraps
            for wrap in wraps:
                if isinstance(data, wrap):
                    return subclass
        return DataTypeTree.subclasses[int]  # For instances created from any custom class.

    def _check_tree_is_correct_one(self, data: object) -> None:
        wraps = self.wraps if isinstance(self.wraps, Iterable) else [self.wraps]
        if not any(isinstance(data, wraps_) for wraps_ in wraps):
            # if type(data) not in wraps:
            wraps_str = [element.__name__ for element in wraps]
            raise DataTypeTreeError(
                f"The given parser ({type(self).__name__}) is meant to parse `{', '.join(wraps_str)}` data type but "
                f"{type(data).__name__} was given"
            )

    def __pre_child_instantiation__(self) -> None:
        ...

    def __post_child_instantiation__(self) -> None:
        ...

    def _get_height(self) -> int:
        """Get maximum height of the current node in the tree."""
        max_height = 0
        for child in self:
            max_height = max(child.height, max_height)
        return max_height + 1

    @staticmethod
    def _validate_name(name: str) -> None:
        """CHeck that any given name is compatible with Python keyword naming rules."""
        if not is_string_python_keyword_compatible(name):
            raise DataTypeTreeError(f"The given name ({name}) is not Python-keyword compatible")

    @abstractmethod
    def _instantiate_children(self, data: object) -> Optional[ChildrenStructure[DataTypeTree]]:
        """Instantiate the child structure that will be assigned to `self.children`.

        This one will depend on how each subclass manage the child structure.
        """

    @final
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Populate `subclasses` with all subclasses of this same class."""
        super().__init_subclass__(**kwargs)
        wraps = cls.wraps if isinstance(cls.wraps, Iterable) else [cls.wraps]
        for type_ in wraps:
            if type_ != object:
                if type_ in cls.subclasses:
                    raise DataTypeTreeError(f"A parser for {type_.__name__} was already found")
                cls.subclasses[type_] = cls  # type: ignore

    @abstractmethod
    def _get_hash(self) -> Hashable:
        """Get a unique hash that identifies the current data type."""

    @final
    def __hash__(self) -> int:
        """Unique hash that identifies whether the current tree is considered to be unique."""
        return hash(self._get_hash())

    @abstractmethod
    def _get_str_top_node(self) -> str:
        """Get the type alias or the representation only for the current self.

        It does not include children.
        """

    @property
    def permission_to_be_created_as_type_alias(self) -> bool:
        """Parameter that indicates whether the current class is allowed to be created as a type alias.

        This will depend on the parsing strategy and on the fact that some trees require a type
        alias no matter its height.
        """
        if self.parent is None:
            return True
        return bool(self.height > self.strategies.min_height_to_define_type_alias)

    @final
    def get_str_top_node_without_lvalue(self) -> str:
        return self.get_str_top_node().split("=")[-1].strip()

    @final
    def get_str_top_node(self) -> str:
        return self._get_str_top_node()

    @final
    def get_str_all_nodes(self, include_imports: bool = True) -> str:
        """String that represents the .py file created from the tree."""
        return self._format_node_strings(self.get_strs_all_nodes_unformatted(include_imports=include_imports))

    @final
    @cache_returned_value
    def get_strs_all_nodes_unformatted(self, *, include_imports: bool = True) -> Tuple[str, ...]:
        """Get, ordered by dependencies, all strings representing the whole tree."""
        strings: OrderedSet[str] = OrderedSet()
        self._get_strs_all_nodes_unformatted(types=strings)
        strings_lst = strings.as_list()
        if include_imports:
            strings_lst.insert(0, self.imports.format())
        return tuple(strings_lst)

    @final
    def _get_strs_all_nodes_unformatted(self, *, types: Optional[OrderedSet[str]] = None) -> None:
        if types is None:
            types = OrderedSet()

        if self.depth == 0 and not self.children:
            if self.permission_to_be_created_as_type_alias:
                types.add(self.get_str_top_node())
            return
        if not self.children:
            return

        for child in self:
            child._get_strs_all_nodes_unformatted(types=types)
            if child.permission_to_be_created_as_type_alias:
                types.add(child.get_str_top_node())
        if self.permission_to_be_created_as_type_alias:
            types.add(self.get_str_top_node())
        return

    @final
    def _format_node_strings(self, strs_py: Sequence[str]) -> str:
        """Get the string representation of the type hints that represent the whole tree."""
        strs_py = list(strs_py)
        if not strs_py:
            return ""

        # Add extra separation between imports and first line
        for idx, line in enumerate(strs_py):  # noqa: B007
            if "" in line:
                break
        strs_py[idx] = strs_py[idx] + "\n"

        # 2 empty lines before defining a class
        for idx, line in enumerate(strs_py[2:], start=2):
            if line.startswith("class"):
                strs_py[idx] = "\n" + strs_py[idx]

        return "\n\n".join(strs_py)

    @final
    def __str__(self) -> str:
        """String that represents the .py file created from the tree."""
        return self.get_str_all_nodes(include_imports=True)

    @final
    def __repr__(self) -> str:
        return f"{type(self).__name__}-{self.name}"

    @abstractmethod
    def __iter__(self) -> Self:
        """Get an iterator to iterate over the children of the tree."""

    @final
    def __len__(self) -> int:
        """Number of children within the tree."""
        if self.children is None:
            return 0
        return len(self.children)

    @abstractmethod
    def __next__(self) -> DataTypeTree:
        """Iterate over the children of the tree."""

    @final
    def print_children(self) -> None:
        print(repr(self))

    @final
    def __eq__(self, other_object: object) -> bool:
        if type(self) is type(other_object):
            return hash(self) == hash(other_object)
        return False



"""Tool that allows to read comments from yaml files and re-introduce them as part of the dictionary."""

import re
import tempfile
from collections import defaultdict
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Final,
    FrozenSet,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

YAML_COMMENTS_POSITION = Literal["above", "below", "side"]
"""Possible locations where comments can be written."""


@dataclass(frozen=True)
class Comment:
    """Characterization of a single multi-line comment."""

    full_string: str
    associated_with_key: str
    key_line: int
    spacing_element: Tuple[Literal["spaces", "tabs", "none"], int]
    must_replace_its_key_as_first_element_of_list: bool = False

    @property
    def spacing(self) -> str:
        spacing_element = "\t" if self.spacing_element[0] == "tabs" else " "
        lst: List[str] = [spacing_element] * self.spacing_element[1]
        if self.must_replace_its_key_as_first_element_of_list:
            lst[-2] = "-"
            return "".join(lst)
        return "".join(lst)


class YamlFileModifierError(Exception):
    ...


class YamlFileModifier:
    path: Path
    """Path where the file lies."""
    lines: Tuple[str, ...]
    """Content of the file parsed as multiple lines."""
    comments_are: Tuple[YAML_COMMENTS_POSITION, ...]
    """Potential location of the comments. Order matters."""
    prefix: Final = "___docstring_hidden_key_"
    """When reading the comments associated with a key, this prefix will be prepended to the key.
    
    This will allow to create a new dictionary mirroring the original one but with new hidden
    keys that capture the comments.
    """

    def __init__(
        self, path: Union[str, Path], *, comments_are: Union[YAML_COMMENTS_POSITION, Sequence[YAML_COMMENTS_POSITION]]
    ) -> None:
        """
        Initialize a YamlFileModifier object.

        Args:
            path (Union[str, Path]): The path to the YAML file.
            comments_are (Union[YAML_COMMENTS_POSITION, Sequence[YAML_COMMENTS_POSITION]]): The position(s) where
                comments are located. If multiple are provided, the order of the sequence will affect how the final
                docstring is built.
        """
        if "above" in comments_are and "below" in comments_are:
            raise YamlFileModifierError(
                f"When choosing where the comments are located, you can choose any combination that does not combine "
                f"`below` and `above` at the same time. Input given: {', '.join(comments_are)}"
            )
        self.path = Path(path)
        if not self.path.suffix.endswith((".yaml", ".yml")):
            raise YamlFileModifierError(f"Only `.yaml` or `.yml` are allowed. File given is: {self.path}")
        self.lines = tuple(Path(path).read_text().splitlines())
        comments_are = (comments_are,) if isinstance(comments_are, str) else tuple(comments_are)
        self.comments_are = comments_are

    @staticmethod
    def _capitalize_only_first_letter(string: str) -> str:
        try:
            return string[0].upper() + string[1:]
        except IndexError:
            return string

    @staticmethod
    def _remove_spacing(line: str) -> str:
        line.strip()
        lines = line.split("\t")
        lines = [line.lstrip() for line in lines]
        return "".join(lines)

    @staticmethod
    def _join_multi_line_comments(lines: Sequence[str]) -> str:
        """
        Joins multiple lines of comments into a single string.

        This one includes the use of `.` and capital letters when needed
        """
        result = ""
        # Remove null strings from both sides of the list until finding first valid string
        lines = list(lines)
        while lines and lines[0] == "":
            lines.pop(0)
        while lines and lines[-1] == "":
            lines.pop()

        if not lines:
            return ""
        for idx, single_line_comment in enumerate(lines):
            if not single_line_comment:
                result += "\n\n"
                continue
            if idx > 0:
                if (
                    single_line_comment and single_line_comment[0].isupper()
                ):  # Check if the string starts with an uppercase letter
                    if result.endswith("."):
                        result += " " + single_line_comment.strip()
                    else:
                        if result.endswith("\n"):
                            single_line_comment_ = single_line_comment.strip()
                            result += YamlFileModifier._capitalize_only_first_letter(single_line_comment_)
                        else:
                            result += ". " + single_line_comment.strip()
                else:
                    single_line_comment_ = single_line_comment.strip()
                    if result.endswith("."):
                        result += " " + YamlFileModifier._capitalize_only_first_letter(single_line_comment_)
                    else:
                        if result.endswith("\n"):
                            single_line_comment_ = single_line_comment.strip()
                            result += YamlFileModifier._capitalize_only_first_letter(single_line_comment_)
                        else:
                            result += " " + single_line_comment.strip()
            else:
                result += single_line_comment.strip()
        if not result:
            return result
        if result.endswith("."):
            return result
        return result + "."

    @staticmethod
    def _extract_side_comment(line: str) -> str:
        """
        Extracts the side comment from a given line of text.

        Example:
            yaml_key: value  # My comment
        """
        idx = YamlFileModifier._find_first_occurrence_that_is_not_between(line, "#")
        if not idx:
            return ""
        try:
            if line[idx - 1] != " ":
                return ""
        except IndexError:
            return ""
        comment = line[idx + 1 :].strip().rstrip()
        if comment and not comment.endswith("."):
            return comment + "."
        return comment

    def _extract_single_block_comment(self, line_idx: int, *, comments_are: Literal["above", "below"]) -> str:
        """
        Extracts a single block comment from the specified line index.

        It is important that the comment is perfectly aligned with the key that is commenting.
        It also merges multiple lines.
        Only comments attached to dictionary keys will be returned.

        Example:
            # Comment above
            key: value
            # Comment below
            # that can be multiple lines

        Args:
            line_idx (int): The index of the line to extract the comment from.
            comments_are (Literal["above", "below"]): Specifies whether the comments are located above
                or below the line.
        """
        # Find first non space/tab character
        try:
            line = self.lines[line_idx]
        except IndexError:
            return ""
        if not self._find_first_occurrence_that_is_not_between(line, ":"):
            return ""
        idx_first_char = line.find(line.strip().replace("-", "").strip()[0])

        step: Final = -1 if comments_are == "above" else 1
        increment = step
        multi_line_comments: List[str] = []
        with suppress(IndexError):
            while self.lines[line_idx + increment][idx_first_char] == "#":
                multi_line_comments.append(self.lines[line_idx + increment][idx_first_char + 1 :].strip())
                increment += step

        if not multi_line_comments:
            return ""

        if comments_are == "above":
            string = self._join_multi_line_comments(multi_line_comments[::-1])
        else:
            string = self._join_multi_line_comments(multi_line_comments)
        return string.replace("'''", "").replace('"""', "")

    @staticmethod
    def _find_first_occurrence_that_is_not_between(
        line: str, occurrence: str, not_between: FrozenSet[str] = frozenset({"'", '"'})
    ) -> Optional[int]:
        """
        Finds the first occurrence of a character that is not between the given characters.

        Args:
            line (str): The line to search for the occurrence.
            occurrence (str): The character to search for.
            not_between (FrozenSet[str], optional): A set of characters that define the quotes.
                Defaults to frozenset({"'", '"'}).

        Returns:
            Optional[int]: The index of the first occurrence that is not between quotes, or None if not found.
        """
        between_quotes = False
        for i, char in enumerate(line):
            if char in not_between:
                between_quotes = not between_quotes
            elif char == occurrence and not between_quotes:
                return i
        return None

    @staticmethod
    def _extract_key_from_line(line: str) -> str:
        """
        Extracts the key from a YAML line.

        Examples for `line`:
            - key: value
              key2: value
        """
        line = YamlFileModifier._remove_spacing(line)
        if line[0] == "-":  # Detect list cases
            line = " " + line[1:]

        idx_first_char = line.find(line.strip()[0])
        idx_last_char = YamlFileModifier._find_first_occurrence_that_is_not_between(line=line, occurrence=":")
        if not idx_last_char:
            return ""
        line = line[idx_first_char:idx_last_char]
        if line.startswith('"') and line.endswith('"'):
            return line.strip('"')
        elif line.startswith("'") and line.endswith("'"):
            return line.strip("'")
        else:
            return line

    @staticmethod
    def _detect_indentation(lines: Sequence[str]) -> Literal["spaces", "tabs", "??"]:
        """
        Detects the type of indentation used in the given lines.

        Args:
            lines (Sequence[str]): The lines of code to analyze.

        Returns:
            Literal["spaces", "tabs", "??"]: The type of indentation used. It can be "spaces" if spaces are
                used for indentation, "tabs" if tabs are used for indentation, or "??" if both spaces and
                tabs are used.

        """
        tabs_used = any("\t" in line for line in lines)
        spaces_used = any(" " in line for line in lines)
        if tabs_used and spaces_used:
            return "??"
        elif tabs_used:
            return "tabs"
        elif spaces_used:
            return "spaces"
        else:
            return "??"

    @staticmethod
    def _count_spaces_or_tabs_at_start(line: str, *, indentation: Literal["spaces", "tabs"]) -> Optional[int]:
        """
        Counts the number of spaces or tabs at the start of a line.

        Args:
            line (str): The line to count spaces or tabs from.
            indentation (Literal["spaces", "tabs"]): The type of indentation to count.

        Returns:
            Optional[int]: The number of spaces or tabs at the start of the line, or None if the line contains a
                hyphen and the indentation is set to "tabs".
        """
        if "-" in line and indentation == "tabs":
            return None
        line = line.replace("-", " ")
        pattern = re.compile(r"^[ \t]*")
        result = pattern.match(line)
        if result:
            spaces_tabs = result.group()
            return len(spaces_tabs)
        return 0

    def _extract_comments(self) -> Tuple[Comment, ...]:
        """
        Extracts comments associated with keys in the YAML file.

        This includes comments that can be located above, below or by the side.
        These comments can also be multi-line.

        Returns:
            A tuple of Comment objects representing the extracted comments.
        """
        comments: List[Comment] = []
        indentation = self._detect_indentation(self.lines)
        if indentation == "??":
            return ()

        for idx, line in enumerate(self.lines):
            line_with_no_spacing = self._remove_spacing(line)
            if len(line_with_no_spacing) > 0 and (line_with_no_spacing[0].isalpha() or line_with_no_spacing[0] == "-"):
                spaces_or_tabs = self._count_spaces_or_tabs_at_start(line, indentation=indentation)
                if spaces_or_tabs is None:
                    continue

                key = self._extract_key_from_line(line)
                if not key:
                    continue

                for comments_are in self.comments_are:
                    if comments_are == "above" or comments_are == "below":
                        comment = self._extract_single_block_comment(idx, comments_are=comments_are)
                        comment = self._capitalize_only_first_letter(comment)
                    elif comments_are == "side":
                        comment = self._extract_side_comment(line)
                    is_key_first_element_within_list = line_with_no_spacing[0] == "-"

                    if not comment:
                        continue
                    # Comments that are either below or above the keyword
                    comments.append(
                        Comment(
                            full_string=comment,
                            key_line=idx,
                            spacing_element=(indentation, spaces_or_tabs),
                            associated_with_key=key,
                            must_replace_its_key_as_first_element_of_list=is_key_first_element_within_list,
                        )
                    )

        return self._merge_comments(comments)

    @staticmethod
    def _format_comment_as_multiline_yaml(comment: Comment) -> str:
        """
        Formats a comment as a multiline YAML string.

        Args:
            comment (Comment): The comment object to format.
        """
        lines = comment.full_string.splitlines()
        lines.insert(0, "")
        indentation = comment.spacing_element[0]
        indent = "\t" if indentation == "tabs" else " "
        multiple_indentation = indent * (comment.spacing_element[1] + 1)
        return f"\n{multiple_indentation}".join(lines)

    @staticmethod
    def _merge_comments(comments: Sequence[Comment]) -> Tuple[Comment, ...]:
        """
        Merge comments that are associated with the same key.

        This can heppen if the user requests to read from "below" and "side" for example.
        """
        comments_in_line: Mapping[int, List[Comment]] = defaultdict(list)
        for comment in comments:
            comments_in_line[comment.key_line].append(comment)

        merged_comments: List[Comment] = []
        for comments_lst in comments_in_line.values():
            string = "\n\n".join(comment.full_string for comment in comments_lst)
            merged_comments.append(
                Comment(
                    full_string=string,
                    key_line=comments_lst[0].key_line,
                    associated_with_key=comments_lst[0].associated_with_key,
                    spacing_element=comments_lst[0].spacing_element,
                    must_replace_its_key_as_first_element_of_list=any(
                        comment.must_replace_its_key_as_first_element_of_list for comment in comments_lst
                    ),
                )
            )
        return tuple(merged_comments)

    def _create_temporary_string_with_comments_as_keys(self) -> str:
        """Create the YAML-based string representation of the new dictionary containing new doc-based keys."""
        comments = self._extract_comments()
        reverse_order_comments = sorted(comments, key=lambda comment: comment.key_line, reverse=True)
        new_lines = list(self.lines).copy()
        for comment in reverse_order_comments:
            key = self.prefix + comment.associated_with_key
            string = f"{comment.spacing}{key}: |{self._format_comment_as_multiline_yaml(comment)}"
            if comment.must_replace_its_key_as_first_element_of_list:
                new_lines[comment.key_line] = new_lines[comment.key_line].replace("-", " ", 1)
            new_lines.insert(comment.key_line, string)
        return "\n".join(new_lines)

    def create_temporary_file_with_comments_as_keys(self) -> Path:
        """Create a new YAML file with new keys containing the documentation.

        This method creates a new YAML file by converting the comments in the original file into keys in the new file.
        The comments are extracted from the original file and added as keys in the new file.

        Returns:
            Path: The path to the newly created YAML file.
        """
        string = self._create_temporary_string_with_comments_as_keys()
        path = Path(tempfile.gettempdir()) / f"_{self.path.name}"
        path.write_text(string)
        return path



from typing import Any, Hashable, Literal, Sequence, Set, Tuple

from typing_extensions import override

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.generic_type.generic_data_type_tree import (
    GenericDataTypeTree,
)
from lazy_type_hint.data_type_tree.generic_type.set_and_sequence_operations import SetAndSequenceOperations


class SetDataTypeTree(GenericDataTypeTree):
    wraps = (frozenset, set)
    children: Sequence[DataTypeTree]
    operations: SetAndSequenceOperations

    @override
    def __pre_child_instantiation__(self) -> None:
        self.operations = SetAndSequenceOperations(self)

    @override
    def _instantiate_children(self, data: Sequence[Any]) -> Tuple[DataTypeTree, ...]:  # type: ignore
        return self.operations.instantiate_children(data, allow_repeated_children=False)

    @override
    def _get_str_top_node(self) -> str:
        container: Literal["FrozenSet", "set"] = "FrozenSet" if self.holding_type is frozenset else "set"
        self.imports.add(container)

        if container == "FrozenSet":
            return f"{self.name} = FrozenSet[{self.get_type_alias_children()}]"
        return f"{self.name} = Set[{self.get_type_alias_children()}]"

    @override
    def _get_hash(self) -> Hashable:
        hashes: Set[object] = set()
        for child in self:
            hashes.add(child._get_hash())
        return frozenset(hashes)



from typing import (
    Dict,
    Final,
    List,
    Mapping,
    Set,
    Type,
)

import pandas as pd
from typing_extensions import override

from lazy_type_hint.data_type_tree import data_type_tree_factory
from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.generic_type.mapping_data_type_tree import MappingDataTypeTree

OVERLOAD_TEMPLATE: Final = """    @overload  # type: ignore
    def __getitem__(self, key: Literal["{literal}"]) -> {rtype}:
        ...
"""
TEMPLATE: Final = """class {class_name}(pd.DataFrame):

{overloads}
    @overload
    def __getitem__(
        self,
        key: Union[
            "pd.Series[bool]",
            {allowed_types},
            pd.DataFrame,
            pd.Index,
            npt.NDArray[np.bool_],
            npt.NDArray[np.str_],
            List[Union[Scalar, Tuple[Hashable, ...]]],
        ],
    ) -> Union[pd.Series, pd.DataFrame]:
        ...

    def __getitem__(
        self,
        key: Union[
            "pd.Series[bool]",
            {allowed_types},
            pd.DataFrame,
            pd.Index,
            npt.NDArray[np.bool_],
            npt.NDArray[np.str_],
            List[Union[Scalar, Tuple[Hashable, ...]]],
        ],
    ) -> Union[pd.Series, pd.DataFrame]:
        return super().__getitem__(key)"""

TEMPLATE_NO_PD: Final = """class {class_name}(pd.DataFrame):

{overloads}
    def __getitem__(
        self,
        key: Union[
            {allowed_types},
            npt.NDArray[np.bool_],
            npt.NDArray[np.str_],
            List[Union[Scalar, Tuple[Hashable, ...]]],
        ],
    ) -> Union[pd.Series, pd.DataFrame]:
        return super().__getitem__(key)"""


class PandasDataFrameDataTypeTree(MappingDataTypeTree):
    wraps = pd.DataFrame
    data: pd.DataFrame
    children: Mapping[str, DataTypeTree]  # type: ignore[assignment]

    @override
    @property
    def permission_to_be_created_as_type_alias(self) -> bool:
        return bool(len(self) >= 1 and self.are_columns_either_all_str_or_all_tuple)

    @property
    def can_be_accessed_multilevel(self) -> bool:
        return self.all_columns_are(tuple)

    @property
    def are_column_same_type(self) -> bool:
        return len({type(column) for column in self.data.columns}) == 1

    def all_columns_are(self, type_: Type[object]) -> bool:
        return all(isinstance(column, type_) for column in self.data.columns)

    @property
    def are_columns_either_all_str_or_all_tuple(self) -> bool:
        if self.are_column_same_type:
            return all(isinstance(column, (str, tuple)) for column in self.data.columns)
        return False

    def _create_child(self, column: str) -> DataTypeTree:
        suffix = self._to_camel_case(column)
        return data_type_tree_factory(  # type: ignore
            data=self.data[column],
            name=f"{self.name}{suffix}",
            imports=self.imports,
            depth=self.depth + 1,
            strategies=self.strategies,
            parent=self,
        )

    @override
    def _instantiate_children(self, data: Mapping[str, object]) -> Mapping[str, DataTypeTree]:  # type: ignore
        children: Dict[str, DataTypeTree] = {}
        if not self.are_columns_either_all_str_or_all_tuple:
            return children

        # Corner case to avoid infinite recursion
        if all(isinstance(column, tuple) and len(column) == 1 for column in self.data.columns):
            return {}

        for column in self.data.columns:
            if not self.can_be_accessed_multilevel:  # Here all columns will  be str
                children[column] = self._create_child(column)
            else:  # Here all columns will be tuple
                columns_processed: Set[str] = set()
                for column in self.data.columns:
                    if column[0] not in columns_processed:
                        columns_processed.add(column[0])
                        children[column[0]] = self._create_child(column[0])
        return children

    @override
    def _get_hash(self) -> str:
        if self.strategies.pandas_strategies == "Do not type hint columns":
            return "pd.DataFrame"
        return str(self.data.columns)

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("pandas")
        if len(self) == 0 or not self.strategies.pandas_strategies:
            self.imports.add("TypeAlias")
            return f"{self.name}: TypeAlias = pd.DataFrame"
        self.imports.add("overload")
        self.imports.add("Union")
        self.imports.add("Literal")
        self.imports.add("npt")
        self.imports.add("tuple")
        self.imports.add("list")
        self.imports.add("Hashable")
        self.imports.add("numpy")
        self.imports.add("pd.Scalar")

        overloads: List[str] = []
        for literal, child in self.children.items():
            if child.permission_to_be_created_as_type_alias:
                overloads.append(OVERLOAD_TEMPLATE.format(literal=literal, rtype=child.name))
            else:
                rtype = "Union[pd.DataFrame, pd.Series]"
                overloads.append(OVERLOAD_TEMPLATE.format(literal=literal, rtype=rtype))
        if self.strategies.pandas_strategies == "Full type hint":
            all_literals = ", ".join(f'"{key}"' for key in self.children)
            allowed_types = f"Literal[{all_literals}]"
            template = TEMPLATE_NO_PD
        else:
            allowed_types = "str"
            template = TEMPLATE
        return template.format(class_name=self.name, overloads="\n".join(overloads), allowed_types=allowed_types)



from contextlib import suppress

from lazy_type_hint.data_type_tree.simple_data_type_tree.function_data_type_tree import (
    FunctionDataTypeTree as FunctionDataTypeTree,
)
from lazy_type_hint.data_type_tree.simple_data_type_tree.instance_data_type_tree import (
    InstanceDataTypeTree as InstanceDataTypeTree,
)

with suppress(ImportError):
    from lazy_type_hint.data_type_tree.simple_data_type_tree.pandas_series_data_type_tree import (
        PandasSeriesDataTypeTree as PandasSeriesDataTypeTree,
    )
from lazy_type_hint.data_type_tree.simple_data_type_tree.io_data_type_tree import IoDataTypeTree as IoDataTypeTree
from lazy_type_hint.data_type_tree.simple_data_type_tree.module_data_type_tree import (
    ModuleTypeDataTypeTree as ModuleTypeDataTypeTree,
)
from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import (
    SimpleDataTypeTree as SimpleDataTypeTree,
)
from lazy_type_hint.data_type_tree.simple_data_type_tree.type_data_type_tree import (
    TypeDataTypeTree as TypeDataTypeTree,
)



from typing import List, Tuple


def _split_string(string: str, *, max_line_length: int) -> Tuple[str, ...]:
    """
    Splits a string into substrings based on a maximum line length.

    Args:
        string (str): The input string to be split.
        max_line_length (int): The maximum length of each line.

    Returns:
        Tuple[str, ...]: A tuple of substrings, where each substring has a length less than or equal to the
            maximum line length.
    """
    substrings: List[str] = []
    for line in string.split("\n"):
        if not line:
            substrings.append("")
        words = line.split()
        current_substring = ""
        for word in words:
            if len(current_substring) + len(word) + 1 <= max_line_length:
                if current_substring:
                    current_substring += " " + word
                else:
                    current_substring = word
            else:
                if current_substring:
                    substrings.append(current_substring)
                current_substring = word
        if current_substring:
            substrings.append(current_substring)
    return tuple(substrings)


def format_string_as_docstring(string: str, *, max_line_length: int = 90, indentation: str = "") -> str:
    """
    Formats a string as a docstring.

    Args:
        string (str): The string to be formatted as a docstring.
        max_line_length (int, optional): The maximum line length for the docstring. Defaults to 90.
        indentation (str, optional): The indentation to be applied to each line of the docstring. Defaults to "".

    Returns:
        str: The formatted docstring.
    """
    string = string.rstrip("\n")
    if not string.endswith("."):
        string += "."
    docstring = '"""' + string + '"""'
    if len(docstring) <= max_line_length and "\n" not in docstring:
        return indentation + docstring
    substrings = _split_string(string, max_line_length=max_line_length)
    if len(substrings) == 1 and "\n" not in docstring:
        return indentation + docstring

    substrings = tuple([f"{indentation}{substring}" for substring in substrings])
    return f'{indentation}"""\n' + "\n".join(substrings) + f'\n{indentation}"""'



from typing import TYPE_CHECKING, Iterator, overload

from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.strategies import ParsingStrategies

if TYPE_CHECKING:
    from types import MappingProxyType, ModuleType
    from typing import (
        Any,
        Dict,
        FrozenSet,
        Iterator,
        List,
        Mapping,
        Optional,
        Sequence,
        Set,
        TextIO,
        Tuple,
        Union,
    )

    import pandas as pd

    from lazy_type_hint.data_type_tree.generic_type.dict_data_type_tree import DictDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.iterator_data_type_tree import IteratorDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.list_data_type_tree import ListDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.mapping_data_type_tree import MappingDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.pandas_data_frame_data_type_tree import PandasDataFrameDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.sequence_data_type_tree import SequenceDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.set_data_type_tree import SetDataTypeTree
    from lazy_type_hint.data_type_tree.generic_type.tuple_data_type_tree import TupleDataTypeTree
    from lazy_type_hint.data_type_tree.simple_data_type_tree.io_data_type_tree import IoDataTypeTree
    from lazy_type_hint.data_type_tree.simple_data_type_tree.module_data_type_tree import ModuleTypeDataTypeTree
    from lazy_type_hint.data_type_tree.simple_data_type_tree.pandas_series_data_type_tree import (
        PandasSeriesDataTypeTree,
    )
    from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree
    from lazy_type_hint.utils import ImportManager


@overload
def data_type_tree_factory(  # type: ignore[overload-overlap]
    data: "Iterator[Any]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "IteratorDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[overload-overlap]
    data: "ModuleType",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "ModuleTypeDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[overload-overlap]
    data: "TextIO",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "IoDataTypeTree":
    ...


@overload
def data_type_tree_factory(
    data: "pd.Series",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "PandasSeriesDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "pd.DataFrame",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "PandasDataFrameDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "List[Any]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "ListDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Union[Set[Any], FrozenSet[Any]]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "SetDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Union[bool, float, slice, None]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "SimpleDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Tuple[Any, ...]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "TupleDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Dict[Any, Any]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "DictDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Sequence[Any]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "SequenceDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: "Union[Mapping[Any, Any], MappingProxyType[Any, Any]]",
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> "MappingDataTypeTree":
    ...


@overload
def data_type_tree_factory(  # type: ignore[misc]
    data: object,
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> DataTypeTree:
    ...


def data_type_tree_factory(
    data: object,
    name: str,
    *,
    imports: "Optional[ImportManager]" = None,
    depth: int = 0,
    strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
    parent: "Optional[DataTypeTree]" = None,
) -> DataTypeTree:
    return DataTypeTree.get_subclass(data)(
        data=data, name=name, imports=imports, depth=depth, strategies=strategies, parent=parent
    )



from lazy_type_hint.generators.lazy_type_hint import LazyTypeHint as LazyTypeHint
from lazy_type_hint.generators.lazy_type_hint_live import LazyTypeHintLive as LazyTypeHintLive
from lazy_type_hint.strategies import ParsingStrategies as ParsingStrategies



from dataclasses import dataclass, field
from typing import Generic, List, Set, Tuple, TypeVar

ObjectT = TypeVar("ObjectT")


@dataclass(frozen=True)
class OrderedSet(Generic[ObjectT]):
    """A data structure that maintains the order of elements while ensuring uniqueness."""

    _elements_added: Set[ObjectT] = field(default_factory=set)
    _lst: List[ObjectT] = field(default_factory=list)

    def add(self, element: ObjectT) -> None:
        """
        Adds an element to the ordered set if it is not already present.

        Args:
            element: The element to be added.
        """
        if element not in self._elements_added:
            self._elements_added.add(element)
            self._lst.append(element)

    def as_tuple(self) -> Tuple[ObjectT, ...]:
        """
        Returns the elements of the ordered set as a tuple.

        Returns:
            Tuple[ObjectT, ...]: The elements of the ordered set as a tuple.
        """
        return tuple(self._lst)

    def as_list(self) -> List[ObjectT]:
        """
        Returns the elements of the ordered set as a list.

        Returns:
            List[ObjectT]: The elements of the ordered set as a list.
        """
        return list(self._lst)



from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree as DataTypeTree
from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTreeError as DataTypeTreeError
from lazy_type_hint.data_type_tree.factory import data_type_tree_factory as data_type_tree_factory
from lazy_type_hint.data_type_tree.generic_type.generic_data_type_tree import GenericDataTypeTree  # noqa: F401
from lazy_type_hint.data_type_tree.simple_data_type_tree import SimpleDataTypeTree  # noqa: F401



import ast
import inspect
import textwrap
from inspect import Parameter
from types import BuiltinFunctionType, FunctionType, MappingProxyType, MethodType
from typing import Any, Callable, Hashable, Optional

from typing_extensions import override

from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree
from lazy_type_hint.utils import TAB


class FunctionDataTypeTree(SimpleDataTypeTree):
    wraps = (FunctionType, staticmethod, classmethod, BuiltinFunctionType, MethodType)
    data: Callable[[Any], Any]

    @property
    def is_lambda(self) -> bool:
        return bool(self.data.__name__ == "<lambda>")

    @override
    @property
    def permission_to_be_created_as_type_alias(self) -> bool:
        if self.is_lambda or self.parent is None:
            return True
        return True

    @property
    def can_be_inspected(self) -> bool:
        try:
            inspect.signature(self.data)
            return True
        except ValueError:
            return False

    def _get_str_top_node(self) -> str:
        if not self.can_be_inspected:
            self.imports.add("Callable")
            return f"{self.name} = Callable"
        if self.is_lambda:
            return self._get_lambda_str()
        return self._get_protocol_str()

    def _get_protocol_str(self) -> str:
        args = str(inspect.signature(self.data))
        self.imports.add("Protocol")
        self.imports.import_all_unkown_symbols_from_signature(args)

        if "->" not in args:
            if not self._has_return():
                return f"""class {self.name}(Protocol):\n{TAB}def __call__{args} -> None: ..."""
            self.imports.add("Any")
            return f"""class {self.name}(Protocol):\n{TAB}def __call__{args} -> Any: ..."""
        return f"""class {self.name}(Protocol):\n{TAB}def __call__{args}: ..."""

    def get_func_params(self) -> MappingProxyType[str, Parameter]:
        signature = inspect.signature(self.data)
        return signature.parameters

    def _get_lambda_str(self) -> str:
        self.imports.add("Callable").add("Any")
        return f"{self.name} = Callable[[{', '.join(['Any']*len(self.get_func_params().values()))}], Any]"

    @override
    def _get_hash(self) -> Hashable:
        if self.is_lambda:
            return str(f"Callable[[{', '.join(['Any']*len(self.get_func_params().values()))}], Any]")
        else:
            return str(inspect.signature(self.data))

    def _has_return(self) -> Optional[bool]:
        code = textwrap.dedent(inspect.getsource(self.data))
        try:
            tree = ast.parse(code)
        except:  # noqa: E722
            return None
        return any(isinstance(node, ast.Return) for node in ast.walk(tree))



import pandas as pd
from typing_extensions import override

from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree


class PandasSeriesDataTypeTree(SimpleDataTypeTree):
    wraps = pd.Series

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("pandas")
        self.imports.add("TypeAlias")
        return f"{self.name}: TypeAlias = pd.Series"



from typing import Any, Hashable, List, Sequence, Tuple

from typing_extensions import override

from lazy_type_hint.data_type_tree import DataTypeTree
from lazy_type_hint.data_type_tree.generic_type.sequence_data_type_tree import SequenceDataTypeTree


class TupleDataTypeTree(SequenceDataTypeTree):
    wraps = tuple

    @override
    def _instantiate_children(self, data: Sequence[Any]) -> Tuple[DataTypeTree, ...]:  # type: ignore
        if self.strategies.tuple_size_strategy == "fixed":
            return self.operations.instantiate_children(data, allow_repeated_children=True)
        else:
            return self.operations.instantiate_children(data, allow_repeated_children=False)

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("tuple")
        return f"{self.name} = Tuple[{self.get_type_alias_children()}]"

    @override
    def get_type_alias_children(self) -> str:
        if self.strategies.tuple_size_strategy == "fixed":
            child_types = self._get_types(remove_repeated=False)
        else:
            child_types = self._get_types(remove_repeated=True)
        return self._format_types(child_types)

    @override
    def _format_types(self, child_types: Sequence[str]) -> str:
        if len(child_types) == 1 and "Any" in child_types:
            self.imports.add("Any")
            return "Any, ..."

        if self.strategies.tuple_size_strategy == "fixed":
            return f"{', '.join(child_types)}"
        else:
            names_set = set(child_types)
            if len(names_set) == 1:
                return f"{next(iter(names_set))}, ..."
            self.imports.add("Union")
            return f"Union[{', '.join(sorted(names_set))}], ..."

    @override
    def _get_hash(self) -> Hashable:
        if self.strategies.tuple_size_strategy == "any size":
            return super()._get_hash()
        else:
            hashes: List[object] = []
            for child in self:
                hashes.append(child._get_hash())
            return tuple(hashes)



import os
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from lazy_type_hint.data_type_tree import data_type_tree_factory
from lazy_type_hint.data_type_tree.data_type_tree import DataTypeTree
from lazy_type_hint.file_modifiers.yaml_file_modifier import YAML_COMMENTS_POSITION, YamlFileModifier
from lazy_type_hint.strategies import ParsingStrategies
from lazy_type_hint.utils import is_string_python_keyword_compatible


class LazyTypeHintError(Exception):
    """Raised by `LazyTypeHint` class."""


PathT = TypeVar("PathT", str, Path)


@dataclass(frozen=True)
class Tree:
    _tree: DataTypeTree

    def to_string(self, *, include_imports: bool = True) -> str:
        return self._tree.get_str_all_nodes(include_imports=include_imports)

    def to_file(self, path_to_py: Union[Path, str], *, create_non_existing_dir: bool = False) -> None:
        path_to_py = Path(path_to_py)
        if not path_to_py.exists():
            if not create_non_existing_dir:
                raise ValueError(
                    "The given directory does not exist and permissions to create the folder "
                    "(create_non_existing_dir input argument) were not set"
                )
            os.makedirs(path_to_py.parent)
        path_to_py.write_text(self.to_string(include_imports=True))


class LazyTypeHintABC(ABC):
    strategies: ParsingStrategies
    """Strategies to follow when parsing the objects."""

    def __init__(
        self,
        strategies: ParsingStrategies = ParsingStrategies(),  # noqa: B008
        **kwargs: Any,
    ) -> None:
        self.strategies = strategies

    def from_yaml_file(
        self,
        loader: Callable[[PathT], object],
        path: PathT,
        *,
        class_name: str,
        comments_are: Optional[Union[YAML_COMMENTS_POSITION, Sequence[YAML_COMMENTS_POSITION]]] = "side",
        **kwargs: Any,
    ) -> Any:
        if comments_are is None:
            return self.from_data(loader(path), class_name=class_name)
        yaml_file_modifier = YamlFileModifier(path, comments_are=comments_are)
        original_data = loader(path)
        try:
            new_path: PathT = type(path)(yaml_file_modifier.create_temporary_file_with_comments_as_keys())
            potentially_modified_data = loader(new_path)
            return self.from_data(potentially_modified_data, class_name=class_name)
        except Exception:  # noqa: BLE001
            return self.from_data(original_data, class_name=class_name)

    def from_data(
        self,
        data: object,
        *,
        class_name: str,
        **kwargs: Any,
    ) -> Any:
        if not is_string_python_keyword_compatible(class_name):
            raise LazyTypeHintError(
                f"Given class_name is not compatible with Python class naming conventions: {class_name}"
            )
        return data_type_tree_factory(data, name=class_name, strategies=self.strategies)



from abc import abstractmethod
from typing import (
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Sequence,
    Tuple,
    Union,
    final,
)

from typing_extensions import Self, override

from lazy_type_hint.data_type_tree.data_type_tree import ChildrenStructure, DataTypeTree


class GenericDataTypeTree(DataTypeTree):
    """Tree that holds any kind of object that must contain other inner structures (children)."""

    children: ChildrenStructure[DataTypeTree]

    # Iterable-protocol related
    _iterator: Union[int, Iterator[Hashable]]

    @abstractmethod
    def _instantiate_children(self, data: object) -> ChildrenStructure[DataTypeTree]:
        ...

    def get_type_alias_children(self) -> str:
        """Get, in a format manner, a single string with all subtypes found within the generic structure.

        Examples:
            - Union[float, MyDict]
            - Union[float, MyDict, str]
            - MyDict
        """
        child_types = self._get_types()
        return self._format_types(child_types)

    def _format_types(self, child_types: Sequence[str]) -> str:
        """Given a collection of type or type aliases, format them.

        If multiple of them are given, a `Union` type will be used.
        """
        if len(child_types) == 1:
            return child_types[0]
        self.imports.add("Union")
        return f"Union[{', '.join(child_types)}]"

    @final
    def _get_types(self, iterable: Iterable[object] = (), *, remove_repeated: bool = True) -> Tuple[str, ...]:
        """
        Get the types of the generic data type tree. Child ones by default.

        This method keeps the order in case the structure used is an ordered one.

        Args:
            iterable (Iterable[object], optional): An iterable of objects. Defaults to an empty iterable.
            remove_repeated (bool, optional): Whether to remove repeated child types. Defaults to True.

        Returns:
            Tuple[str, ...]: A tuple containing the child types.
        """
        child_types: List[str] = []
        if iterable:
            child_types = [type(element).__name__ for element in iterable]
        else:
            for child in self:
                if not child.permission_to_be_created_as_type_alias:
                    child_types.append(child.get_str_top_node_without_lvalue())
                else:
                    child_types.append(child.name)
        if not child_types:
            self.imports.add("Any")
            child_types.append("Any")

        if remove_repeated:
            child_types_set = sorted(set(child_types))
            if "float" in child_types and "int" in child_types:
                child_types_set.remove("int")
            return tuple(sorted(child_types_set))
        return tuple(child_types)

    def __iter__(self) -> "Self":
        if isinstance(self.children, Mapping):
            self._iterator = iter(self.children.keys())
        else:
            self._iterator = 0  # Reset the index to zero when starting a new iteration
        return self

    @override
    def __next__(self) -> DataTypeTree:
        # Many type ignores here but everything is safe
        if isinstance(self.children, Mapping):
            key = next(self._iterator, None)  # type: ignore
            if key is not None:
                return self.children[key]
            else:
                raise StopIteration
        else:
            if self._iterator < len(self.children):  # type: ignore
                element = list(self.children)[self._iterator]  # type: ignore
                self._iterator += 1  # type: ignore
                return element
            else:
                raise StopIteration

    @override
    def _get_hash(self) -> Hashable:
        hashes: List[object] = []
        for child in self:
            hashes.append(child._get_hash())
        return tuple(hashes)



import io

from typing_extensions import override

from lazy_type_hint.data_type_tree.simple_data_type_tree.simple_data_type_tree import SimpleDataTypeTree


class IoDataTypeTree(SimpleDataTypeTree):
    wraps = io.IOBase

    @override
    def _get_str_top_node(self) -> str:
        self.imports.add("TextIO")
        return f"{self.name} = TextIO"
